# -*- coding: utf-8 -*-

from logilab.common.testlib import TestCase, unittest_main
from StringIO import StringIO

from pybill.lib.config.xmlreaders import load_config
from pybill.lib.errors import PyBillReadingException

xml_data = u"""
<pbc:config xmlns:pbc="http://www.logilab.org/2010/PyBillConfig" pbc:format-version="PBC-1.0" pbc:name="Z0">
 <pbc:company>
  <pbc:logo-file>A1</pbc:logo-file>
  <pbc:orgname>A2</pbc:orgname>
  <pbc:address>
   <pbc:street>A3</pbc:street>
   <pbc:street>A4</pbc:street>
   <pbc:pob>A5</pbc:pob>
   <pbc:postcode>A6</pbc:postcode>
   <pbc:city>A7</pbc:city>
   <pbc:state>A8</pbc:state>
   <pbc:country>A9</pbc:country>
   <pbc:phone>A10</pbc:phone>
   <pbc:fax>A11</pbc:fax>
   <pbc:web>A12</pbc:web>
   <pbc:email>A13</pbc:email>
  </pbc:address>
 </pbc:company>

 <pbc:bank-data>
  <pbc:line>B1</pbc:line>
  <pbc:line>B2</pbc:line>
 </pbc:bank-data>

 <pbc:agreement-intro>
  <pbc:line>C1</pbc:line>
  <pbc:line>C2</pbc:line>
 </pbc:agreement-intro>

 <pbc:footer>
  <pbc:line>D1</pbc:line>
  <pbc:line>D2</pbc:line>
 </pbc:footer>

 <pbc:number-separators>
  <pbc:sign>F1</pbc:sign>
  <pbc:thousands>F2</pbc:thousands>
  <pbc:digits>F3</pbc:digits>
 </pbc:number-separators>

 <pbc:localisation>
  <pbc:colon>E0</pbc:colon>
  <pbc:phone-kw>E1</pbc:phone-kw>
  <pbc:fax-kw>E2</pbc:fax-kw>
  <pbc:web-kw>E3</pbc:web-kw>
  <pbc:email-kw>E4</pbc:email-kw>
  <pbc:doc-ref-kw>E5</pbc:doc-ref-kw>
  <pbc:receiver-kw>E9</pbc:receiver-kw>
  <pbc:sender-kw>E10</pbc:sender-kw>
  <pbc:on-date>E11</pbc:on-date>
  <pbc:bill>E12</pbc:bill>
  <pbc:claim-form>E13</pbc:claim-form>
  <pbc:downpayment>E14</pbc:downpayment>
  <pbc:pro-forma>E15</pbc:pro-forma>
  <pbc:debit>E16</pbc:debit>
  <pbc:number>E17</pbc:number>
  <pbc:dated>E18</pbc:dated>
  <pbc:valid-until>E19</pbc:valid-until>
  <pbc:intro-detail>E20</pbc:intro-detail>
  <pbc:quantity>E21</pbc:quantity>
  <pbc:description>E22</pbc:description>
  <pbc:vat-rate>E23</pbc:vat-rate>
  <pbc:unit-price>E24</pbc:unit-price>
  <pbc:tf-unit-price>E25</pbc:tf-unit-price>
  <pbc:price>E26</pbc:price>
  <pbc:tf-price>E27</pbc:tf-price>
  <pbc:it-price>E28</pbc:it-price>
  <pbc:holdback-on>E29</pbc:holdback-on>
  <pbc:ita-est>E30</pbc:ita-est>
  <pbc:total>E31</pbc:total>
  <pbc:tf-total>E32</pbc:tf-total>
  <pbc:vat-amount>E33</pbc:vat-amount>
  <pbc:it-total>E34</pbc:it-total>
  <pbc:including>E35</pbc:including>
  <pbc:holdback>E36</pbc:holdback>
  <pbc:on-tf>E37</pbc:on-tf>
  <pbc:on-vat>E38</pbc:on-vat>
  <pbc:debit-total>E39</pbc:debit-total>
  <pbc:charged-downpayment>E40</pbc:charged-downpayment>
  <pbc:charged-on>E41</pbc:charged-on>
  <pbc:issued-debit>E42</pbc:issued-debit>
  <pbc:issued-on>E43</pbc:issued-on>
  <pbc:to-be-paid>E44</pbc:to-be-paid>
  <pbc:payment-terms>E45</pbc:payment-terms>
  <pbc:to-bring-forward>E46</pbc:to-bring-forward>
  <pbc:carry-forward>E47</pbc:carry-forward>
  <!-- Kept for backward compatibility with format of pybill version 0.X -->
  <pbc:your-ref-kw>E101</pbc:your-ref-kw>
  <pbc:purch-ref-kw>E102</pbc:purch-ref-kw>
  <pbc:supplier-ref-kw>E103</pbc:supplier-ref-kw>
 </pbc:localisation>

</pbc:config>
"""

xml_olddata = u"""
<config >
 <company>
  <logo-file>A1</logo-file>
  <orgname>A2</orgname>
  <address>
   <street>A3</street>
   <street>A4</street>
   <pob>A5</pob>
   <postcode>A6</postcode>
   <city>A7</city>
   <state>A8</state>
   <country>A9</country>
   <phone>A10</phone>
   <fax>A11</fax>
   <web>A12</web>
   <email>A13</email>
  </address>
 </company>

 <bank-data>
  <line>B1</line>
  <line>B2</line>
 </bank-data>

 <agreement-intro>
  <line>C1</line>
  <line>C2</line>
 </agreement-intro>

 <footer>
  <line>D1</line>
  <line>D2</line>
 </footer>

 <localisation>
  <colon>E0</colon>
  <phone-kw>E1</phone-kw>
  <fax-kw>E2</fax-kw>
  <web-kw>E3</web-kw>
 </localisation>

</config>
"""

local_dict = { u'colon'              : u"E0",
               u'phone-kw'           : u"E1",
               u'fax-kw'             : u"E2",
               u'web-kw'             : u"E3",
               u'email-kw'           : u"E4",
               u'doc-ref-kw'         : u"E5",
               u'receiver-kw'        : u"E9",
               u'sender-kw'          : u"E10",
               u'on-date'            : u"E11",
               u'bill'               : u"E12",
               u'claim-form'         : u"E13",
               u'downpayment'        : u"E14",
               u'pro-forma'          : u"E15",
               u'debit'              : u"E16",
               u'number'             : u"E17",
               u'dated'              : u"E18",
               u'valid-until'        : u"E19",
               u'intro-detail'       : u"E20",
               u'quantity'           : u"E21",
               u'description'        : u"E22",
               u'vat-rate'           : u"E23",
               u'unit-price'         : u"E24",
               u'tf-unit-price'      : u"E25",
               u'price'              : u"E26",
               u'tf-price'           : u"E27",
               u'it-price'           : u"E28",
               u'holdback-on'        : u"E29",
               u'ita-est'            : u"E30",
               u'total'              : u"E31",
               u'tf-total'           : u"E32",
               u'vat-amount'         : u"E33",
               u'it-total'           : u"E34",
               u'including'          : u"E35",
               u'holdback'           : u"E36",
               u'on-tf'              : u"E37",
               u'on-vat'             : u"E38",
               u'debit-total'        : u"E39",
               u'charged-downpayment': u"E40",
               u'charged-on'         : u"E41",
               u'issued-debit'       : u"E42",
               u'issued-on'          : u"E43",
               u'to-be-paid'         : u"E44",
               u'payment-terms'      : u"E45",
               u'to-bring-forward'   : u"E46",
               u'carry-forward'      : u"E47",
               u'your-ref-kw'        : u"E101",
               u'purch-ref-kw'       : u"E102",
               u'supplier-ref-kw'    : u"E103",
             }


class ConfigXMLReaderTest(TestCase):
    def test_load_unknown_file(self):
        """
        Tests the XMLConfigReader raises an exception when trying to read
        an unknown XML file.
        """
        self.assertRaises(PyBillReadingException,
                          load_config, u"unknown.xml")

    def test_load_wrong_file(self):
        """
        Tests the XMLConfigReader raises an exception when trying to read
        a file that is not a configuration file.
        """
        wrong_data = u"<wrong/>"
        # uses the XML parser sympathy to pass it a string stream instead of a
        # filename
        stream = StringIO(wrong_data)
        self.assertRaises(PyBillReadingException,
                          load_config, stream)

    def test_load_wrong_version_file(self):
        """
        Tests the XMLConfigReader raises an exception when trying to read
        a file that is not in a recognized format.
        """
        wrong_data = u'<pbc:config xmlns:pbc="http://www.logilab.org/2010/PyBillConfig" pbc:format-version="PBD-1.1"/>'
        # uses the XML parser sympathy to pass it a string stream instead of a
        # filename
        stream = StringIO(wrong_data)
        self.assertRaises(PyBillReadingException,
                          load_config, stream)

    def test_load_data(self):
        """
        Tests the XMLConfigReader reads correctly an XML config file
        """
        # uses the XML parser sympathy to pass it a string stream instead of a
        # filename
        stream = StringIO(xml_data)
        cfg = load_config(stream)
        self.assertEqual(cfg.reference, None)
        self.assertEqual(cfg.name, u"Z0")
        self.assertEqual(cfg.company_logo, u"A1")
        self._check_cfg_company_address(cfg.company_address)
        self.assertEqual(cfg.bank_data_lines, [u"B1", u"B2"])
        self.assertEqual(cfg.agreement_intro_lines, [u"C1", u"C2"])
        self.assertEqual(cfg.footer_lines, [u"D1", u"D2"])
        self.assertEqual(cfg.number_separators, {u"sign": u"F1",
                                                  u"thousands": u"F2",
                                                  u"digits": u"F3"})
        self.assertEqual(cfg.local_phrases, local_dict)

    def test_load_olddata(self):
        """
        Tests the XMLConfigReader reads correctly an XML config file in previous
        format
        """
        # uses the XML parser sympathy to pass it a string stream instead of a
        # filename
        stream = StringIO(xml_olddata)
        cfg = load_config(stream)
        self.assertEqual(cfg.reference, None)
        self.assertEqual(cfg.name, u"")
        self.assertEqual(cfg.company_logo, u"A1")
        self._check_cfg_company_address(cfg.company_address)
        self.assertEqual(cfg.bank_data_lines, [u"B1", u"B2"])
        self.assertEqual(cfg.agreement_intro_lines, [u"C1", u"C2"])
        self.assertEqual(cfg.footer_lines, [u"D1", u"D2"])
        self.assertEqual(cfg.number_separators, {})
        self.assertEqual(cfg.local_phrases,
                          {u'colon'   : u"E0",
                           u'phone-kw': u"E1",
                           u'fax-kw'  : u"E2",
                           u'web-kw'  : u"E3",})

    def _check_cfg_company_address(self, addr):
        """
        Checks the company address of the config data contains expected data
        """
        self.assertEqual(addr.name, u"A2")
        self.assertEqual(addr.streets, [u"A3", u"A4"])
        self.assertEqual(addr.post_box, u"A5")
        self.assertEqual(addr.post_code, u"A6")
        self.assertEqual(addr.city, u"A7")
        self.assertEqual(addr.state, u"A8")
        self.assertEqual(addr.country, u"A9")
        self.assertEqual(addr.phone, u"A10")
        self.assertEqual(addr.fax, u"A11")
        self.assertEqual(addr.web, u"A12")
        self.assertEqual(addr.email, u"A13")
        self.assertEqual(addr.divisions, [])
        self.assertEqual(addr.job_titles, [])


# definitions for automatic unit testing

if __name__ == '__main__':
    unittest_main()



# -*- coding: utf-8 -*-
"""
``pybill.main`` is a module that contains the main program of PyBill.

Basically, this program just contains the correct run function that will be run
by the binary scripts. Currently, this run function is the
:func:`~pybill.commandline.ui.run` function from :mod:`pybill.commandline.ui`
module.

This module can be executed and just runs the run function described above.
"""
__docformat__ = "restructuredtext en"

from pybill.commandline.ui import run

if __name__ == '__main__':
    run()



# -*- coding: utf-8 -*-

from logilab.common.testlib import TestCase, unittest_main

import datetime
from pybill.lib.entities.accounting_docs import (Bill, ClaimForm, Debit,
                                                 Downpayment, ProForma)
from pybill.lib.entities.accounting_docs.utils import (AccItem,
                                                       PreviousAccountingDoc)
from pybill.lib.config.entities import ConfigData
from pybill.lib.accounting.utils import (Entry, CLIENT, CLIENT_HOLDBACK,
                                         PRODUCT, VAT)
from pybill.lib.accounting.generator import EntriesGenerator


class EntriesGeneratorTest(TestCase):
    def setUp(self):
        """
        Called before each test from this class
        """
        self.generator = EntriesGenerator()
        self.acc_doc = None

    def test_empty_object(self):
        self.assertEqual(self.generator.entries, [])

    def test_reset(self):
        self.generator.entries = [
            Entry(datetime.date(2009, 11, 06), u"ENTRY 1"),
            Entry(datetime.date(2009, 11, 06), u"ENTRY 2"),
                                 ]
        self.generator.reset_entries()
        self.assertEqual(self.generator.entries, [])

    def test_entry_generation_for_bill_without_holdbacks_previous_docs(self):
        self.acc_doc = Bill("id1")
        self._fill_accounting_doc()
        self.generator.generate_entry(self.acc_doc)
        self._check_entry(u"Bill # id1",
                          {u"4111": [1139.18]},
                          {u"701": [1050.00], u"4451": [89.18]})

    def test_entry_generation_for_bill_without_holdbacks(self):
        self.acc_doc = Bill("id1")
        self._fill_accounting_doc()
        for i in range(2):
            pad = PreviousAccountingDoc()
            pad.total = 2.0 * (i+1)
            pad.vat_amount = 0.7 * i
            self.acc_doc.charged_downpayments.append(pad)
        for i in range(2):
            pad = PreviousAccountingDoc()
            pad.total = 3.0 * (i+1)
            pad.vat_amount = 0.5 * i
            self.acc_doc.issued_debits.append(pad)
        self.generator.generate_entry(self.acc_doc)
        self._check_entry(u"Bill # id1",
                          {u"4111": [1124.18]},
                          {u"701": [1036.20], u"4451": [87.98]})

    def test_entry_generation_for_bill(self):
        self.acc_doc = Bill("id1")
        self._fill_accounting_doc(with_holdbacks=True)
        for i in range(2):
            pad = PreviousAccountingDoc()
            pad.total = 2.0 * (i+1)
            pad.vat_amount = 0.7 * i
            self.acc_doc.charged_downpayments.append(pad)
        for i in range(2):
            pad = PreviousAccountingDoc()
            pad.total = 3.0 * (i+1)
            pad.vat_amount = 0.5 * i
            self.acc_doc.issued_debits.append(pad)
        self.generator.generate_entry(self.acc_doc)
        self._check_entry(u"Bill # id1",
                          {u"4111": [1072.06], u"4112": [52.12]},
                          {u"701": [1036.20], u"4451": [87.98]})

    def test_entry_generation_for_claimform_without_holdbacks(self):
        self.acc_doc = ClaimForm("id1")
        self._fill_accounting_doc()
        self.generator.generate_entry(self.acc_doc)
        self._check_entry(u"Claim Form # id1",
                          {u"4111": [1139.18]},
                          {u"701": [1050.00], u"4451": [89.18]})

    def test_entry_generation_for_claimform(self):
        self.acc_doc = ClaimForm("id1")
        self._fill_accounting_doc(with_holdbacks=True)
        self.generator.generate_entry(self.acc_doc)
        self._check_entry(u"Claim Form # id1",
                          {u"4111": [1087.06], u"4112": [52.12]},
                          {u"701": [1050.00], u"4451": [89.18]})

    def test_entry_generation_for_debit_without_holdbacks(self):
        self.acc_doc = Debit("id1")
        self._fill_accounting_doc()
        self.generator.generate_entry(self.acc_doc)
        self._check_entry(u"Debit # id1",
                          {u"701": [1050.00], u"4451": [89.18]},
                          {u"4111": [1139.18]})

    def test_entry_generation_for_debit(self):
        self.acc_doc = Debit("id1")
        self._fill_accounting_doc(with_holdbacks=True)
        self.generator.generate_entry(self.acc_doc)
        self._check_entry(u"Debit # id1",
                          {u"701": [1050.00], u"4451": [89.18]},
                          {u"4111": [1087.06], u"4112": [52.12]})

    def test_entry_generation_for_downpayment_without_holdbacks(self):
        self.acc_doc = Downpayment("id1")
        self._fill_accounting_doc()
        self.generator.generate_entry(self.acc_doc)
        self._check_entry(u"Downpayment # id1",
                          {u"4111": [341.75]},
                          {u"701": [315.00], u"4451": [26.75]})

    def test_entry_generation_for_downpayment(self):
        self.acc_doc = Downpayment("id1")
        self._fill_accounting_doc(with_holdbacks=True)
        self.generator.generate_entry(self.acc_doc)
        self._check_entry(u"Downpayment # id1",
                          {u"4111": [341.75]},
                          {u"701": [315.00], u"4451": [26.75]})

    def test_entry_generation_for_downpayment_with_huge_percent(self):
        self.acc_doc = Downpayment("id1")
        self.acc_doc.percent = 98.0
        self._fill_accounting_doc(with_holdbacks=True)
        self.generator.generate_entry(self.acc_doc)
        self._check_entry(u"Downpayment # id1",
                          {u"4111": [1087.06], u"4112": [29.34]},
                          {u"701": [1029.00], u"4451": [87.40]})

    def test_entry_generation_for_proforma_without_holdbacks(self):
        self.acc_doc = ProForma()
        self._fill_accounting_doc()
        self.generator.generate_entry(self.acc_doc)
        self.assertEqual(len(self.generator.entries), 0)

    def test_entry_generation_for_proforma(self):
        self.acc_doc = ProForma()
        self._fill_accounting_doc(with_holdbacks=True)
        self.generator.generate_entry(self.acc_doc)
        self.assertEqual(len(self.generator.entries), 0)

    def _check_entry(self, label, exp_debits, exp_credits):
        """
        Checks the unique accounting entry in the generator list.
        ``exp_debits`` and ``exp_credits`` are the expected debits and credits
        of this entry.
        """
        self.assertEqual(len(self.generator.entries), 1)
        entry = self.generator.entries[0]
        self.assertEqual(entry.date, datetime.date(2009, 11, 06))
        self.assertEqual(entry.explanation, label)
        self.assertEqual(entry.debits, exp_debits)
        self.assertEqual(entry.credits, exp_credits)
        
    def _fill_accounting_doc(self, with_holdbacks=False):
        """
        Fills the accounting document (``self.acc_doc``) used for the tests
        with some data.
        """
        self.acc_doc.date_num = datetime.date(2009, 11, 06)
        self.acc_doc.account_numbers = {CLIENT: u"4111",
                                        CLIENT_HOLDBACK: u"4112",
                                        VAT: u"4451", PRODUCT: u"701"}
        for i in range(4):
            item = AccItem()
            item.quantity = 5.0*(i+1)
            item.unit_price = 7.0*(i+1)
            if i in [1,2]:
                item.vat_rate = 19.6
            if with_holdbacks and i in [1,2,3]:
                item.holdback_rate = 5.0
                if i == 1:
                    item.holdback_on_vat = True
            self.acc_doc.acc_items.append(item)
        cfg = ConfigData()
        cfg.local_phrases = {u"bill": u"Bill",
                             u"claim-form": u"Claim Form",
                             u"debit": u"Debit",
                             u"downpayment": u"Downpayment",
                             u"pro-forma": u"Pro-Forma",
                             u"number": u"#"}
        self.acc_doc.cfg_data = cfg


# definitions for automatic unit testing

if __name__ == '__main__':
    unittest_main()



# -*- coding: utf-8 -*-

from logilab.common.testlib import TestCase, unittest_main

from pybill.lib.pdfwriters import PDFWriter
from pybill.lib.entities.accounting_docs import Debit

from generic_test_pdfwriter import PDFWriterGenericTest


class PDFWriterDebitTest(TestCase, PDFWriterGenericTest):
    def setUp(self):
        """
        Called before each test from this class.
        """
        self.writer = PDFWriter()
        self.acc_doc = Debit("identifier")
        self.acc_doc.cfg_data = self._build_config_data()        
        self._fill_acc_doc_metadata(self.acc_doc)
        self.acc_doc.sender = self._build_person_address(u"sd")
        self.acc_doc.receiver = self._build_person_address(u"rc")
        self.acc_doc.payment_terms = "payment_terms"
        self.entity_name = "debit"


# definitions for automatic unit testing

if __name__ == '__main__':
    unittest_main()



# -*- coding: utf-8 -*-
"""
``pybill.lib.xmlreaders.utils`` module defines the public class used for reading
accounting XML files ans building the corresponding entity objects.

The XML files are read with :mod:`lxml` library. They contain an accounting
document that can be a bill, a claim form, a debit, a downpayment or a
pro-forma. The information read from these files is stored in ``Bill``,
``ClaimForm``, ``Debit``, ``Downpayment`` or ``ProForma`` objects.

.. autoclass:: AccDocXMLReader
   :members:
   :undoc-members:
   :show-inheritance:
"""
__docformat__ = "restructuredtext en"

from traceback import format_exc
from lxml import etree
from logilab.common.xmlutils import parse_pi_data

from pybill.lib import PBD_0_X, PBD_1_0 
from pybill.lib.accounting import CLIENT, CLIENT_HOLDBACK, PRODUCT, VAT
from pybill.lib import PBD_XMLNS
from pybill.lib.errors import PyBillReadingException

from pybill.lib.xmlreaders.accdoc_format_1_0 import AccDocFormat_1_0_Reader
from pybill.lib.xmlreaders.accdoc_format_0_X import AccDocFormat_0_X_Reader


class AccDocXMLReader:
    """
    Class reading an XML file containing an accounting document.

    After opening the XML file thanks to `lxml` XML parser, this class
    finds a processing instruction that specifies the name of the configuration
    file to be used for turning the accounting document in PDF, instanciates a
    specialized reader object matching the format of the accounting document,
    and asks it to read the data. Then, it returns the entity object built by
    the specialized reader. This entity object is, eventually, linked to the
    configuration object.

    :ivar format_readers:
        Dictionary of specialized format readers that can read a given format of
        accounting document; these readers are indexed by the name of the format
        they can read. This dictionary contains all the specialized format
        readers supported by this reader.

    :type format_readers:
        dictionary of
        :class:`~pybill.lib.xmlreaders.utils.AccDocFormatAbstractReader` indexed
        by :class:`unicode` keys.

    :ivar cfg_register:
        Configuration register that returns a
        :class:`~pybill.lib.config.entities.ConfigData` object matching a
        configuration name. This register searches and reads the configuration
        file. If this file is not found, returns a default configuration
        object. The configuration is used here to convert accounting documents
        from older formats.

    :type cfg_register: :class:`~pybill.lib.config.register:ConfigRegister`

    .. automethod:: __init__
    """
    def __init__(self, config_register):
        """
        Initializes a new ``AccDocXMLReader`` object.

        :parameter config_register:
            Register containing the various configurations read by PyBill. This
            register is used to get the configuration used by the accounting
            documents read by this object (the configurations can be useful to
            convert data from older formats).
                                    
        :type config_register:
            :class:`~pybill.lib.config.register.ConfigRegister`
        """
        self.format_readers = {
            PBD_1_0: AccDocFormat_1_0_Reader(),
            PBD_0_X: AccDocFormat_0_X_Reader(),
                              }
        self.cfg_register = config_register

    def load_data(self, xml_file, config_ref=None):
        """
        Loads the accounting document contained in the file ``xml_file``.

        If the accounting document contains a pybill processing instruction
        giving the name of the configuration file to be used for document
        transformation in PDF, reads it and linkks it to the accounting entity.
        
        :parameter xml_file: name of the XML file containing the accounting
                             document to be read.
        :type xml_file: :class:`unicode`
        
        :parameter config_ref: reference of the XML configuration file that
                               has been manually specified and that will
                               override any configuration specification in the
                               accounting document.
        :type config_ref: :class:`unicode`
        
        :returns: the entity object containing the accounting document. This
                  object contains also a link towards the configuration object
                  that shall be used for its further process.
        :rtype: :class:`~pybill.lib.entities.accounting_docs.abstract.GenericAccountingDoc`
        """
        try:
            xml_tree = etree.parse(xml_file)
        except Exception, exc:
            raise PyBillReadingException("Error while reading accounting "
                                         "document from file %s\n%s"
                                         % (xml_file, format_exc(limit=1)))
        if config_ref is None:
            # Tries to find the reference of the configuration in a `pybill`
            # processing instruction.
            xml_pis = xml_tree.xpath(u"//processing-instruction('pybill')")
            for xml_pi in xml_pis:
                data = parse_pi_data(xml_pi.text)
                if data.has_key(u"config"):
                    config_ref = unicode(data[u"config"])
                    break
        cfg = self.cfg_register.get_config(config_ref)
        # Tries to find information about accounting accounts in a dedicated
        # processing instruction
        xml_pis = xml_tree.xpath(u"//processing-instruction('accounts')")
        accounts = {}
        if len(xml_pis) > 0:
            data = parse_pi_data(xml_pis[0].text)
            if data.has_key(u"client") and data[u"client"].strip() != u"":
                accounts[CLIENT] = data[u"client"].strip()
            if (data.has_key(u"client-holdback")
                and data[u"client-holdback"].strip() != u""):
                accounts[CLIENT_HOLDBACK] = data[u"client-holdback"].strip()
            if data.has_key(u"product") and data[u"product"].strip() != u"":
                accounts[PRODUCT] = data[u"product"].strip()
            if data.has_key(u"vat") and data[u"vat"].strip() != u"":
                accounts[VAT] = data[u"vat"].strip()
        # Reads the beginning of the document to guess the format and use
        # the proper specialized reader
        xml_root = xml_tree.getroot()
        doc_format = xml_root.get(u"{%s}format-version" % PBD_XMLNS, None)
        if doc_format is None:
            doc_format = xml_root.get(u"format-version", None)
        if (doc_format is None
            and xml_root.tag in [u"bill", u"claim-form", u"debit",
                                 u"downpayment", u"pro-forma"]):
            doc_format = PBD_0_X
        try:
            reader = self.format_readers[doc_format]
        except KeyError:
            raise PyBillReadingException("Document from %s file is not in a "
                                         "PyBill known format" % xml_file)
        # Reads the accounting document data, thanks to the proper reader.
        acc_doc = reader.read_accounting_doc(xml_root, xml_file, cfg)
        acc_doc.cfg_data = cfg
        acc_doc.account_numbers = accounts
        return acc_doc



# -*- coding: utf-8 -*-
"""
``pybill.lib.xmlreaders.utils`` module defines various classes and functions
used in the different specialized format readers.

It also defines the base class for the format readers, or functions to
read the addresses defined in an XML element inspired by the DocBook standard,
or dump adress objects into XML elements.

.. autoclass:: AccDocFormatAbstractReader
   :show-inheritance:

.. autofunction:: read_person_address

.. autofunction:: fill_address_from_address_block

.. autofunction:: read_text_from_xml_elt
"""
__docformat__ = "restructuredtext en"

from traceback import format_exc
from pybill.lib import PBD_XMLNS
from pybill.lib.errors import PyBillReadingException
from pybill.lib.entities.addresses import PersonAddress, OrganisationAddress
from datetime import date
import re


RE_DATE = re.compile("([\d]{4})-([\d]{2})-([\d]{2})[.]*")


class AccDocFormatAbstractReader:
    """
    Base class for the various specialized format readers. A format reader can
    read the XML data in a given format and build the corresponding accounting
    document.

    .. attribute:: filename

       Attribute containing the name of the file that the XML data is read
       from. This file name is mostly used when reporting errors.

       type: :class:`unicode`

    .. automethod:: __init__

    .. automethod:: read_accounting_doc

    .. automethod:: _select_from_xpaths

    .. automethod:: _read_number_from_xpaths

    .. automethod:: _read_date_from_xpaths

    .. automethod:: _read_text_from_xpaths
    """
    def __init__(self):
        """
        Initializes a new ``AccDocFormatAbstractReader`` object.

        As this class is abstract, raises an exception if the object class is
        not one of the child classes.
        """
        if self.__class__ is AccDocFormatAbstractReader:
            raise NotImplementedError("Abtract classes can't be instanciated")
        self.filename = u""

    def read_accounting_doc(self, xml_root, filename="", cfg=None):
        """
        Reads the data contained in the XML under the element
        ``xml_root`` and builds an accounting entity object.

        The class of the accounting entity object (``Bill``, ``ClaimForm``,
        ``Debit``, ``Downpayment``, ``ProForma``) depends on the data read in
        the XML.

        This abstract polymorphic method will be implemented in the child
        classes.
        
        :parameter xml_root: root element of the XML tree containing the
                             accounting document. The XML has already been
                             read with `lxml` parser.
        :type xml_root: :class:`lxml.etree.Element`
        
        :parameter filename: name of the file where the XML root element was
                             read from. This name is used to display nicest
                             errors.
        :type filename: :class:`unicode`

        :parameter cfg: Configuration object that might be useful for
                        format conversion. This object contains the
                        configuration read from a configuration XML file.
        :type cfg: :class:`~pybill.lib.config.entities.ConfigData`
        
        :returns: entity object containing the accounting document.
        :rtype: :class:`~pybill.lib.entities.accounting_docs.abstract.GenericAccountingDoc`
        """
        raise NotImplementedError()

    def _select_from_xpaths(self, xml_elt, xpaths):
        """
        Utilitary function that selects XML elements from an XPath pointer.

        More precisely, the XPath pointer is a list of several XPath that will
        be tried in the order of the list until one of them selects something.
        If none of them selects something, an empty result is returned.

        In the XPaths, you can use the ``pbd:`` prefix that is automatically
        binded to the PyBill document XML namespace (defined in
        :const:`PBD_XMLNS`).

        The result of the selection is returned as a list of XML nodes.

        :parameter xml_elt: XML element where the selection will be started
                            from (especially if the XPath are relative paths).
                            This element has been read in a file thanks
                            to `lxml` parser.
        :type xml_elt: :class:`lxml.etree.Element`

        :parameter xpaths: List of relative XPaths that points to the piece of
                           information that must be selected in the XML
                           element. If nothing is selected by the first XPath,
                           tries the second one, and so on until something is
                           selected.
        :type xpaths: :list of class:`unicode`

        :returns: List of XML nodes (element, attribute, text, etc.) selected
                  by one of the XPaths. If no piece of information was found,
                  returns an empty list.
        :rtype: list of :class:`object`
        """
        selection = []
        for xpath in xpaths:
            selection = xml_elt.xpath(xpath, namespaces={"pbd": PBD_XMLNS})
            if len(selection) > 0:
                break
        return selection

    def _read_number_from_xpaths(self, xml_elt, xpaths, converted_type,
                                 def_value=None, infoname=u""):
        """
        Utilitary function that reads a piece of information in an XML element
        and tries to convert it to a float or an integer.

        If the information doesn't exist, returns a default value. If several
        pieces of information exist, uses the first one. If there is an error
        during the numeric conversion, returns a nice error message.

        In the XPaths, you can use the ``pbd:`` prefix that is automatically
        binded to the PyBill document XML namespace (defined in
        :const:`PBD_XMLNS`).

        :parameter xml_elt: XML element where the piece of information will be
                            read. This element has been read in a file thanks
                            to `lxml` parser.
        :type xml_elt: :class:`lxml.etree.Element`

        :parameter xpaths: List of relative XPaths that points to the piece of
                           information that must be read in the XML element. If
                           nothing is selected by the first XPath, tries the
                           second one, and so on until something is selected.
        :type xpaths: :list of class:`unicode`

        :parameter converted_type: Type of number that the information must be
                                   converted to. Can be ``float`` or ``int``
        :type converted_type: :class:`type`

        :parameter def_value: Default value that is returned if the piece of
                              information can't be read in the XML element.
        :type def_value: :class:`float`

        :parameter infoname: Name of the information that must be read in the
                             XML element. Only used to display nice error
                             messages.
        :type infoname: :class:`unicode`

        :returns: Piece of information read in the XML element and converted to
                  a float or an integer. If no piece of information was found,
                  returns the default value.
        :rtype: :class:`float`
        """
        xml_info = self._select_from_xpaths(xml_elt, xpaths)
        if len(xml_info) == 0:
            return def_value
        try:
            info = converted_type(xml_info[0])
        except ValueError:
            if converted_type is float:
                str_type = u"float"
            elif converted_type is int:
                str_type = u"int"
            else:
                str_type = str(converted_type)
            raise PyBillReadingException("Error while reading %s file, when "
                                         "trying to convert %s specification "
                                         "('%s') to %s."
                                         % (self.filename, infoname,
                                            unicode(xml_info[0]), str_type))
        return info

    def _read_date_from_xpaths(self, xml_elt, xpaths, infoname=u""):
        """
        Utilitary function that reads a piece of information in an XML element
        and tries to convert it to a date.

        If the information doesn't exist, returns ``None``. If several pieces
        of information exist, uses the first one. If there is an error in the
        conversion, returns a nice error message.

        In the XPaths, you can use the ``pbd:`` prefix that is automatically
        binded to the PyBill document XML namespace (defined in
        :const:`PBD_XMLNS`).

        :parameter xml_elt: XML element where the piece of information will be
                            read. This element has been read in a file thanks
                            to `lxml` parser.
        :type xml_elt: :class:`lxml.etree.Element`

        :parameter xpaths: List of relative XPaths that points to the piece of
                           information that must be read in the XML element. If
                           nothing is selected by the first XPath, tries the
                           second one, and so on until something is selected.
        :type xpaths: :list of class:`unicode`

        :parameter infoname: Name of the information that must be read in the
                             XML element. Only used to display nice error
                             messages.
        :type infoname: :class:`unicode`

        :returns: Piece of information read in the XML element and converted to
                  a date object. If no piece of information was found,
                  returns ``None``.
        :rtype: :class:`datetime.date`
        """
        xml_info = self._select_from_xpaths(xml_elt, xpaths)
        if len(xml_info) == 0:
            return None
        info = xml_info[0]
        try:
            assert RE_DATE.match(info)
            year, month, day = RE_DATE.match(info).groups()
            date_obj = date(int(year), int(month), int(day))
        except (AssertionError, ValueError), exc:
            msg = ("Error while reading %s file, when trying to convert %s "
                   "specification ('%s') to a date. Dates should be like "
                   "'CCYY-MM-DD'."
                   % (self.filename, infoname, unicode(info)))
            if exc.__class__ is ValueError:
                msg += "\n%s" % format_exc(limit=1)
            raise PyBillReadingException(msg)
        return date_obj

    def _read_text_from_xpaths(self, xml_elt, xpaths):
        """
        Utilitary function that reads a piece of information in an XML element
        and returns it as an unicode string.

        If the information doesn't exist, returns ``u""``. If several pieces of
        information exist, uses the first one.

        In the XPaths, you can use the ``pbd:`` prefix that is automatically
        binded to the PyBill document XML namespace (defined in
        :const:`PBD_XMLNS`).

        :parameter xml_elt: XML element where the piece of information will be
                            read. This element has been read in a file thanks
                            to `lxml` parser.
        :type xml_elt: :class:`lxml.etree.Element`

        :parameter xpaths: List of relative XPaths that points to the piece of
                           information that must be read in the XML element. If
                           nothing is selected by the first XPath, tries the
                           second one, and so on until something is selected.
        :type xpaths: :list of class:`unicode`

        :returns: Piece of information read in the XML element as an unicode
                  string. If no piece of information was found,
                  returns ``u""``.
        :rtype: :class:`unicode`
        """
        xml_info = self._select_from_xpaths(xml_elt, xpaths)
        if len(xml_info) == 0:
            return u""
        return unicode(xml_info[0])
 

#### Utilitary functions


def read_person_address(xml_addr, xml_ns=u""):
    """
    Utilitary function that reads an address defined in an XML element (\
    ``xml_addr``\ ) and stores it in an entity object.

    Hence, the XML defining an address was inspired by DocBook format and is the
    same in several PyBill formats and in the configuration files.

    :parameter xml_addr: XML element containing the definition of an
                         address in a DocBook-like format. This element
                         was read thanks to `lxml` parser.
    :type xml_addr: :class:`lxml.etree.Element`

    :parameter xml_ns: XML namespace where the XML address elements are
                       defined. Default value is ``u""`` (no namespace).
    :type xml_ns: :class:`unicode`

    :returns: An entity object storing the address information read from
              the XML element.
    :rtype: :class:`~pybill.lib.entities.addresses.PersonAddress`
    """
    addr = PersonAddress()
    addr.honorific = read_text_from_xml_elt(xml_addr,
                                             u"{%s}honorific" % xml_ns)
    addr.firstname = read_text_from_xml_elt(xml_addr, 
                                             u"{%s}firstname" % xml_ns)
    addr.other_name = read_text_from_xml_elt(xml_addr, 
                                              u"{%s}othername" % xml_ns)
    addr.surname = read_text_from_xml_elt(xml_addr, u"{%s}surname" % xml_ns)
    addr.lineage = read_text_from_xml_elt(xml_addr, u"{%s}lineage" % xml_ns)
    fill_address_from_address_block(addr, xml_addr, xml_ns)
    xml_aff = xml_addr.find(u"{%s}affiliation" % xml_ns)
    if xml_aff is not None:
        addr.organisation = OrganisationAddress()
        org = addr.organisation
        for jobtit in xml_aff.findall(u"{%s}jobtitle" % xml_ns):
            line = read_text_from_xml_elt(jobtit, u".")
            if line:
                org.job_titles.append(line)
        org.name = read_text_from_xml_elt(xml_aff, u"{%s}orgname" % xml_ns)
        for orgdiv in xml_aff.findall(u"{%s}orgdiv" % xml_ns):
            line = read_text_from_xml_elt(orgdiv, u".")
            if line:
                org.divisions.append(line)
        xml_org_addr = xml_aff.find(u"{%s}address" % xml_ns)
        if xml_org_addr is not None:
            fill_address_from_address_block(org, xml_org_addr, xml_ns)
    return addr


def fill_address_from_address_block(addr, xml_addr, xml_ns=u""):
    """
    Reads data from an address XML element and fills it in an address
    object.

    The XML defining an address was inspired by DocBook format.

    :parameter addr: An entity object that will be filled with the address
                     information read from the XML element. This object
                     can be a person address or an organisation address. It
                     is modified by this function.
    :type addr: :class:`~pybill.lib.entities.addresses.Address`

    :parameter xml_addr: XML element containing the definition of an
                         address in a DocBook-like format. This element
                         was read thanks to `lxml` parser.
    :type xml_addr: :class:`lxml.etree.Element`

    :parameter xml_ns: XML namespace where the XML address elements are
                       defined. Default value is ``u""`` (no namespace).
    :type xml_ns: :class:`unicode`
    """
    for street in xml_addr.findall(u"{%s}street" % xml_ns):
        line = read_text_from_xml_elt(street, u".")
        if line:
            addr.streets.append(line)
    addr.post_box = read_text_from_xml_elt(xml_addr, u"{%s}pob" % xml_ns)
    addr.post_code = read_text_from_xml_elt(xml_addr, u"{%s}postcode" % xml_ns)
    addr.city = read_text_from_xml_elt(xml_addr, u"{%s}city" % xml_ns)
    addr.state = read_text_from_xml_elt(xml_addr, u"{%s}state" % xml_ns)
    addr.country = read_text_from_xml_elt(xml_addr, u"{%s}country" % xml_ns)
    addr.phone = read_text_from_xml_elt(xml_addr, u"{%s}phone" % xml_ns)
    addr.fax = read_text_from_xml_elt(xml_addr, u"{%s}fax" % xml_ns)
    addr.web = read_text_from_xml_elt(xml_addr, u"{%s}web" % xml_ns)
    addr.email = read_text_from_xml_elt(xml_addr, u"{%s}email" % xml_ns)


def read_text_from_xml_elt(xml_startpoint, elt_name, def_value=u""):
    """
    Utilitary private function that returns the text contained in an
    XML element named ``elt_name`` selected from an XML start point.

    This function is mainly used by the functions reading the addresses from
    XML elements.

    :parameter xml_startpoint: XML element where the selection starts from.
    :type xml_addr: :class:`lxml.etree.Element`

    :parameter elt_name: Name of the XML element (most of the times, a 
                         child of the XML startpoint) containing the text we 
                         want to get. The name is given in ``lxml`` format:
                         ``{namespace}elt_name/{namespace}elt_name``.
    :type xpath: :class:`unicode`

    :parameter def_value: Default value to be returned if no XML element is
                          selected with ``elt_name``.
    :type def_value: :class:`unicode`

    :returns: The text contained in the XML element selected by ``elt_name`` 
              from ``xml_startpoint`` element. If the XML element is empty, 
              returns ``u""``. If the XML element doesn't exist, returns
              ``def_value``.
    :rtype: :class:`unicode`
    """
    xml_elt = xml_startpoint.find(elt_name)
    if xml_elt is None:
        return def_value
    text = xml_elt.findtext(u".")
    if text is None:
        return u""
    else:
        return unicode(text)



# -*- coding: utf-8 -*-
"""
``pybill.lib.entities.accounting_docs`` is a sub-package that defines the
entities modelizing the accounting documents.

These entities represent bills, claim forms, debits, downpayments, pro-formas,
and some inner business objects.

The sub-package is divided in several modules, each defining some of the entity
classes.

``utils`` module
================

.. automodule:: pybill.lib.entities.accounting_docs.utils

``abstract`` module
===================

.. automodule:: pybill.lib.entities.accounting_docs.abstract

``bill`` module
===============

.. automodule:: pybill.lib.entities.accounting_docs.bill

``claimform`` module
====================

.. automodule:: pybill.lib.entities.accounting_docs.claimform

``debit`` module
================

.. automodule:: pybill.lib.entities.accounting_docs.debit

``downpayment`` module
======================

.. automodule:: pybill.lib.entities.accounting_docs.downpayment

``proforma`` module
===================

.. automodule:: pybill.lib.entities.accounting_docs.proforma
"""
__docformat__ = "restructuredtext en"

# Imports the entity classes exported by this package.

from pybill.lib.entities.accounting_docs.bill import Bill
from pybill.lib.entities.accounting_docs.claimform import ClaimForm
from pybill.lib.entities.accounting_docs.debit import Debit
from pybill.lib.entities.accounting_docs.downpayment import Downpayment
from pybill.lib.entities.accounting_docs.proforma import ProForma

from pybill.lib.entities.accounting_docs.utils import AccItem
from pybill.lib.entities.accounting_docs.utils import PreviousAccountingDoc



# -*- coding: utf-8 -*-
"""
``pybill.lib.accounting`` package contains the classes used to generate the
accounting entries from the accounting documents processed by PyBill.

The accounting entries can be saved in an XML file in `pycompta` format. This
package uses :mod:`lxml` to generate and write the XML data.

The only public class that should be used outside of this package is the
:class:`~pybill.lib.accounting.generator.EntriesGenerator` class. This class is
imported and available in the namespace of this package. Some useful constants
used to designate account numbers are also imported and available in the
namespace of this package.
"""
__docformat__ = "restructuredtext en"


# Imports exported classes and constants.

from pybill.lib.accounting.utils import CLIENT, CLIENT_HOLDBACK, PRODUCT, VAT
from pybill.lib.accounting.generator import EntriesGenerator



# -*- coding: utf-8 -*-

from logilab.common.testlib import TestCase, unittest_main

from pybill.lib.entities.addresses import OrganisationAddress
from pybill.lib.config.entities import ConfigData


class ConfigDataTest(TestCase):

    def test_instanciation(self):
        """
        Tests the instanciation of an empty object.
        """
        cfg = ConfigData()
        self.assertEqual(cfg.reference, None)
        self.assertEqual(cfg.name, u"")
        self.assertEqual(cfg.company_logo, u"")
        self.assertTrue(cfg.company_address.__class__ is OrganisationAddress)
        # Just tests that the object seems to be empty
        self.assertEqual(cfg.company_address.name, u"")
        self.assertEqual(cfg.footer_lines, [])
        self.assertEqual(cfg.bank_data_lines, [])
        self.assertEqual(cfg.agreement_intro_lines, [])
        self.assertEqual(cfg.number_separators, {})
        self.assertEqual(cfg.local_phrases, {})

    def test_get_local_with_existing_keyword(self):
        """
        Tests the ``get_local_phrase`` with an existing keyword.
        """
        cfg = ConfigData()
        cfg.local_phrases = {u"A1": u"B1", u"A2": u"B2", u"A3": u"B3"}
        self.assertEqual(cfg.get_local_phrase(u"A2"), u"B2")
        self.assertEqual(cfg.get_local_phrase(u"A3"), u"B3")

    def test_get_local_with_unknown_keyword(self):
        """
        Tests the ``get_local_phrase`` with a non-existing keyword.
        """
        cfg = ConfigData()
        cfg.local_phrases = {u"A1": u"B1", u"A2": u"B2", u"A3": u"B3"}
        self.assertEqual(cfg.get_local_phrase(u"A4"), u"## A4 ##")
        self.assertEqual(cfg.get_local_phrase(u"B1"), u"## B1 ##")


# definitions for automatic unit testing

if __name__ == '__main__':
    unittest_main()




# -*- coding: utf-8 -*-
"""
``pybill.lib.errors`` module contains the specific exceptions that can be raised
during a PyBill run.

.. autoexception:: PyBillException
   :members:
   :undoc-members:
   :show-inheritance: 

.. autoexception:: PyBillReadingException
   :members:
   :undoc-members:
   :show-inheritance: 

.. autoexception:: PyBillWritingException
   :members:
   :undoc-members:
   :show-inheritance: 

.. autoexception:: PyBillProcessingException
   :members:
   :undoc-members:
   :show-inheritance: 
"""
__docformat__ = "restructuredtext en"


class PyBillException(Exception):
    """
    Base class of the exceptions raised by PyBill.
    """
    pass

class PyBillReadingException(PyBillException):
    """
    Exception raised when an error occurs during reading.
    """
    pass

class PyBillWritingException(PyBillException):
    """
    Exception raised when an error occurs during writing.
    """
    pass

class PyBillProcessingException(PyBillException):
    """
    Exception raised when an error occurs during processing.
    """
    pass



# -*- coding: utf-8 -*-

from StringIO import StringIO
import reportlab 

from pybill.lib.entities.addresses import PersonAddress, OrganisationAddress
from pybill.lib.entities.accounting_docs import AccItem
from pybill.lib.config.entities import ConfigData

# When set to True, PDF results are written on the disk in a subdirectory of
# data directory. This is the way to check manually the PDF produced by the
# tests.
WRITE_RESULTS_ON_DISK = False

REFERENCE_DIR = ("reference_pdfs/reportlab_%s"
                 % reportlab.Version.replace(".", "_"))


class PDFWriterGenericTest :
    """
    Abstract class containing various useful methods for testing the PDF
    generation for an accounting document and factorizing some common tests.

    We should not define test methods on this object as it doesn't inherit from
    ``TestCase``... but it is so much simpler to do this and Python is a nice
    dynamic language that doesn't complain about this.

    :ivar writer: PDFWriter object that will be built by the ``setUp``
                  method before each test and that will be used to transform
                  the accounting document in PDF.
    :type writer: `pybill.lib.pdfwriters.PDFWriter`
    
    :ivar acc_doc: Accounting document that will be built by the ``setUp``
                   method before each test and that will be transformed in
                   PDF in each test.
    :type acc_doc: `pybill.lib.entities.accounting_docs.GenericAccountingDoc`

    :ivar entity_name: Name of the accounting entity that is tested. This name
                       is used to generate the filenames of the PDF (output PDF
                       and reference PDF). It will be defined in the child
                       classes by the ``setUp`` method, run before each test.
                       Can be "bill", "claimform", "debit", "downpayment",
                       or "proforma".
    :type entity_name: str
    """
    def __init__(self) :
        self.writer = None
        self.acc_doc = None
        self.entity_name = ""
        if self.__class__ is PDFWriterGenericTest :
            raise NotImplementedError()

    def _mask_pdf_variable_parts(self, pdf_str):
        """
        This method deletes the variable parts in a PDF string. Hence, in a PDF
        some keyword contain the date of creation, the checksum, etc. In order
        to compare a PDF with a PDF that was previously manually validated, it
        is interesting to delete these parts.

        Currently, this method deletes:

        - the literal strings after ``CreationDate`` and ``ModDate`` keywords,

        - the hexadecimal string after ``CheckSum`` keyword,

        - the trailer of the PDF document (that contains an unique identifier
          for the document).

        Please note that this method doesn't modify the initial PDF string but
        returns a modified copy of this string.

        :parameter pdf_str: Unicode string containing the whole PDF document.
                            This string is copied and some parts of the PDF
                            document are masked.
        :type pdf_str: unicode

        :returns: Unicode string containing an extract of the initial PDF
                  document where some parts were deleted. This string is a copy
                  of the initial string. This string doesn't contain a valid
                  PDF document.
        :rtype: unicode
        """
        masked_str = pdf_str[:]
        # Masks creation date, modification date and checksum.
        for keyword, start_tag, end_tag in [("CreationDate", "(", ")"),
                                            ("ModDate", "(", ")"),
                                            ("CheckSum", "<", ">")]:
            while (True):
                try:
                    kw_start = masked_str.index("/%s" %keyword)
                except ValueError:
                    break
                # Just to make sure we don't go too far in case of problem.
                dict_end = masked_str.find(">>", kw_start)
                val_start = min(masked_str.index(start_tag, kw_start),
                                dict_end)
                val_end = min(masked_str.index(end_tag, val_start),
                              dict_end)
                masked_str = masked_str[:kw_start] + masked_str[val_end+1:]
        # Masks document trailer.
        start = masked_str.rfind("trailer") + 7
        end = masked_str.find("startxref", start)
        masked_str = masked_str[:start] + "\n" + masked_str[end:]
        return masked_str
    
    def _run_pdf_test(self, pdf_filename):
        """
        Runs the common part of a PDF test: builds the PDF with ``self.writer``
        from ``self.acc_doc``. If ``WRITE_RESULTS_ON_DISK`` constant is set to
        ``True``, the resulting PDF will be written in a subdirectory of the
        data directory.

        The resulting PDF is compared with a reference PDF that has been
        manually validated. Please note that this comparison is really basic:
        two document might be different but have the same rendering in a PDF
        viewer. Here, we just check that the two documents are strictly
        identical. However, as PDF document have some variable parts (date of
        creation, etc.), we mask these parts before the actual comparison.

        Please be aware that the PDF documents are stored in strings (that are
        copied for doing the masking). Therefore, it is unwise to test large
        PDF documents.

        :parameter pdf_filename: Name of the PDF files: the output PDF that
                                 would be written in ``output_pdfs`` data
                                 subdirectory if ``WRITE_RESULTS_ON_DISK``
                                 constant is set, and the reference PDF that
                                 is read from ``reference_pdfs`` data
                                 subdirectory.
        :type pdf_filename: str
        """
        # Uses the reportlab `platypus` ``BaseDocTemplate`` sympathy to pass
        # it a string stream instead of a filename.
        output_pdf = StringIO()
        self.writer.write(self.acc_doc, output_pdf)
        output_pdf.flush()
        # Writes the resulting PDF on the disk if wanted.
        if WRITE_RESULTS_ON_DISK:
            output_file = open(self.datapath("output_pdfs/%s" % pdf_filename),
                               'w')
            output_file.write(output_pdf.getvalue())
            output_file.close()
        # Tries to load the reference PDF.
        try:
            ref_pdf = open(self.datapath("%s/%s"
                                         % (REFERENCE_DIR, pdf_filename)),
                           'r').read()
        except IOError:
            ref_pdf = ""
        # Masks the variable parts in the PDFs and compares them
        out = self._mask_pdf_variable_parts(output_pdf.getvalue())
        ref = self._mask_pdf_variable_parts(ref_pdf)
        self.assertEqual(out, ref)

    def _fill_acc_doc_metadata(self, acc_doc) :
        """
        Fills the metadata for ``acc_doc`` accounting document.

        :parameter acc_doc: Accounting document whose common metadata fields
                            will be filled. This object is modified by this
                            method.
        :type acc_doc:
            `pybill.lib.entities.accounting_docs.GenericAccountingDoc`
        """
        acc_doc.doc_ref = u"document reference"
        acc_doc.place = u"place"
        acc_doc.date = u"date"
        acc_doc.other_infos = {u"keyword1": u"value1",
                               u"keyword2": u"value2"}

    def _build_person_address(self, code) :
        """
        Builds an ``PersonAddress`` and fills it with data.

        :parameter code: String prefix that will be added in all the fields
                         of the address and that eases the future checks.
        :type code: unicode

        :returns: A ``PersonAddress`` filled with data.
        :rtype: `pybill.lib.entities.addresses.PersonAddress`
        """
        addr = PersonAddress()
        addr.honorific = u"%s_hon" % code
        addr.firstname = u"%s_firstname" % code
        addr.other_name = u"%s_oth" % code
        addr.surname = u"%s_surname" % code
        addr.lineage = u"%s_lin" % code
        addr.streets.append(u"%s_street_1" % code)
        addr.streets.append(u"%s_street_2" % code)
        addr.post_box = u"%s_pb" % code
        addr.post_code = u"%s_pcode" % code
        addr.city = u"%s_city" % code
        addr.state = u"%s_state" % code
        addr.country = u"%s_country" % code
        addr.phone = u"%s_phone" % code
        addr.fax = u"%s_fax" % code
        addr.email = u"%s_email" % code
        addr.web = u"%s_web" % code
        # affiliation
        addr.organisation = self._build_organisation_address(u"%s_affil" \
                                                             % code)
        return addr

    def _build_organisation_address(self, code) :
        """
        Builds an ``OrganisationAddress`` and fills it with data.

        :parameter code: String prefix that will be added in all the fields
                         of the address and that eases the future checks.
        :type code: unicode

        :returns: An ``OrganisationAddress`` filled with data.
        :rtype: `pybill.lib.entities.addresses.OrganisationAddress`
        """
        addr = OrganisationAddress()
        addr.name = u"%s_name" % code
        addr.job_titles.append(u"%s_job_title_1" % code)
        addr.job_titles.append(u"%s_job_title_2" % code)
        addr.divisions.append(u"%s_division_1" % code)
        addr.divisions.append(u"%s_division_2" % code)
        addr.streets.append(u"%s_street_1" % code)
        addr.streets.append(u"%s_street_2" % code)
        addr.post_box = u"%s_pb" % code
        addr.post_code = u"%s_pcode" % code
        addr.city = u"%s_city" % code
        addr.state = u"%s_state" % code
        addr.country = u"%s_country" % code
        addr.phone = u"%s_phone" % code
        addr.fax = u"%s_fax" % code
        addr.email = u"%s_email" % code
        addr.web = u"%s_web" % code
        return addr

    def _build_acc_items_list(self, length) :
        """
        Builds an accounting items list of length ``length``, filled with
        different ``AccItem`` objects.

        :parameter length: Length of the list of accounting items to be built.
        :type length: int
        
        :returns: A list of ``AccItem`` objects whose length is ``length``.
        :rtype: list
        """
        acclist = []
        for i in range(length):
            it = AccItem()
            it.quantity = float(i+1)
            it.title = u"Element %d" % (i+1)
            it.details.append(u"detailed description %d.1" % (i+1))
            it.details.append(u"detailed description %d.2" % (i+1))
            it.unit_price = float(10*(i+1.007453))
            it.vat_rate = 19.6
            it.holdback_rate = None
            it.holdback_on_vat = False
            acclist.append(it)
        return acclist

    def _build_config_data(self) :
        """
        Builds and fills a ``ConfigData`` object.

        :returns: A ``ConfigData`` object filled with relevant information.
        :rtype: `pybill.lib.config.entities.ConfigData`
        """
        cfg = ConfigData()
        cfg.company_address = self._build_organisation_address(u"cp")
        cfg.footer_lines.append(u"footer: line 1")
        cfg.footer_lines.append(u"footer: line 2")
        cfg.bank_data_lines.append(u"bank data: line 1")
        cfg.bank_data_lines.append(u"bank data: line 2")
        cfg.agreement_intro_lines.append(u"agreement introduction: line 1")
        cfg.agreement_intro_lines.append(u"agreement introduction: line 2")
        cfg.number_separators = {u"sign"     : u" ",
                                 u"thousands": u",",
                                 u"digits"   : u"·"}
        cfg.local_phrases = {u'colon'              : u":",
                             u'phone-kw'           : u"Phone",
                             u'fax-kw'             : u"Fax",
                             u'web-kw'             : u"Web",
                             u'email-kw'           : u"Email",
                             u'doc-ref-kw'         : u"Our ref",
                             u'receiver-kw'        : u"Attn",
                             u'sender-kw'          : u"Followed by",
                             u'on-date'            : u"on",
                             u'bill'               : u"Bill",
                             u'claim-form'         : u"Claim Form",
                             u'downpayment'        : u"Downpayment",
                             u'pro-forma'          : u"Purchase Order",
                             u'debit'              : u"Debit",
                             u'number'             : u"# ",
                             u'dated'              : u"dated",
                             u'valid-until'        : u"valid until",
                             u'intro-detail'       : u"All amounts in Euros",
                             u'quantity'           : u"Qty",
                             u'description'        : u"Description",
                             u'vat-rate'           : u"VAT rate",
                             u'unit-price'         : u"Unit Price",
                             u'tf-unit-price'      : u"TF Unit Price",
                             u'price'              : u"Price",
                             u'tf-price'           : u"TF Price",
                             u'it-price'           : u"IT Price",
                             u'holdback-on'        : u"Holdback on",
                             u'ita-est'            : u"i.e.",
                             u'total'              : u"Total",
                             u'tf-total'           : u"TF Total",
                             u'vat-amount'         : u"VAT",
                             u'it-total'           : u"IT Total",
                             u'including'          : u"including",
                             u'holdback'           : u"Holdback",
                             u'on-tf'              : u"on TF amounts",
                             u'on-vat'             : u"on VAT amounts",
                             u'debit-total'        : u"Total of debit",
                             u'charged-downpayment': u"Downpayment",
                             u'charged-on'         : u"charged on",
                             u'issued-debit'       : u"Debit",
                             u'issued-on'          : u"issued on",
                             u'to-be-paid'         : u"To be paid (EUR)",
                             u'payment-terms'      : u"Payment Terms",
                             u'to-bring-forward'   : u"To bring forward",
                             u'carry-forward'      : u"Carry forward",
                             u'your-ref-kw'        : u"Your ref",
                             u'purch-ref-kw'       : u"Purchase #",
                             u'supplier-ref-kw'    : u"Supplier #",
                            }

        return cfg

    def test_write_one_item(self):
        """
        Tests the PDFWriter writes correctly an accounting document with only
        one item.
        """
        self.acc_doc.acc_items = self._build_acc_items_list(1)
        self.acc_doc.acc_items[0].details = []
        self._run_pdf_test("%s_one_item.pdf" % self.entity_name),

    def test_write_one_item_with_details(self):
        """
        Tests the PDFWriter writes correctly an accounting document with only
        one item with details.
        """
        self.acc_doc.acc_items = self._build_acc_items_list(1)
        self.acc_doc.acc_items[0].details[0] = u"detailed description 1.1" * 7
        self._run_pdf_test("%s_one_item_with_details.pdf" % self.entity_name)

    def test_write_three_items(self):
        """
        Tests the PDFWriter writes correctly an accounting document with three
        items.
        """
        self.acc_doc.acc_items = self._build_acc_items_list(3)
        self._run_pdf_test("%s_three_items.pdf" % self.entity_name)

    def test_write_three_items_with_no_sender(self):
        """
        Tests the PDFWriter writes correctly an accounting document with three
        items and no sender data.
        """
        self.acc_doc.acc_items = self._build_acc_items_list(3)
        self.acc_doc.sender = None
        self._run_pdf_test("%s_three_items_no_sender.pdf" % self.entity_name)

    def test_write_three_items_with_person_receiver(self):
        """
        Tests the PDFWriter writes correctly an accounting document with three
        items and a receiver that is a person (no organisation data).
        """
        self.acc_doc.acc_items = self._build_acc_items_list(3)
        self.acc_doc.receiver.organisation = None
        self._run_pdf_test("%s_three_items_person_receiver.pdf" 
                           % self.entity_name)

    def test_write_three_items_with_organisation_receiver(self):
        """
        Tests the PDFWriter writes correctly an accounting document with three
        items and a receiver that is an organisation (no person name).
        """
        self.acc_doc.acc_items = self._build_acc_items_list(3)
        self.acc_doc.receiver.honorific = u""
        self.acc_doc.receiver.firstname = u""
        self.acc_doc.receiver.other_name = u""
        self.acc_doc.receiver.surname = u""
        self.acc_doc.receiver.lineage = u""
        self._run_pdf_test("%s_three_items_organisation_receiver.pdf" 
                           % self.entity_name)

    def test_write_three_items_with_various_qty_digits(self):
        """
        Tests the PDFWriter writes correctly an accounting document with three
        items and different specifications of quantity digits.
        """
        self.acc_doc.acc_items = self._build_acc_items_list(3)
        for i in range(3):
            self.acc_doc.acc_items[i].quantity = 1.545
            self.acc_doc.acc_items[i].quantity_digits = i
        self._run_pdf_test("%s_three_items_with_various_qty_digits.pdf"
                           % self.entity_name)

    def test_write_three_items_with_various_price_digits(self):
        """
        Tests the PDFWriter writes correctly an accounting document with three
        items and different specifications of unit price digits.
        """
        self.acc_doc.acc_items = self._build_acc_items_list(3)
        for i in range(3):
            self.acc_doc.acc_items[i].unit_price_digits = i+2
        self._run_pdf_test("%s_three_items_with_various_price_digits.pdf"
                           % self.entity_name)

    def test_write_three_items_with_various_vat_rates(self):
        """
        Tests the PDFWriter writes correctly an accounting document with three
        items that have various VAT rates.
        """
        self.acc_doc.acc_items = self._build_acc_items_list(3)
        self.acc_doc.acc_items[1].vat_rate = 33.0
        self._run_pdf_test("%s_three_items_with_various_vat_rates.pdf"
                           % self.entity_name)

    def test_write_three_items_with_various_vat_rates_and_one_nul(self):
        """
        Tests the PDFWriter writes correctly an accounting document with three
        items that have various VAT rates, one of this rate is zero.
        """
        self.acc_doc.acc_items = self._build_acc_items_list(3)
        self.acc_doc.acc_items[1].vat_rate = 33.0
        self.acc_doc.acc_items[2].vat_rate = 0.0
        self._run_pdf_test("%s_three_items_with_various_vat_rates_and_one"
                           "_nul.pdf" % self.entity_name)

    def test_write_three_items_with_various_vat_rates_one_not_defined(self):
        """
        Tests the PDFWriter writes correctly an accounting document with three
        items that have various VAT rates, one of this rate is not defined.
        """
        self.acc_doc.acc_items = self._build_acc_items_list(3)
        self.acc_doc.acc_items[1].vat_rate = 33.0
        self.acc_doc.acc_items[2].vat_rate = None
        self._run_pdf_test("%s_three_items_with_various_vat_rates_one_not"
                           "_defined.pdf" % self.entity_name)

    def test_write_three_items_with_no_vat_rate(self):
        """
        Tests the PDFWriter writes correctly an accounting document with three
        items that have no VAT rate.
        """
        self.acc_doc.acc_items = self._build_acc_items_list(3)
        for i in range(3):
            self.acc_doc.acc_items[i].vat_rate = None
        self._run_pdf_test("%s_three_items_with_no_vat_rate.pdf"
                           % self.entity_name)

    def test_write_three_items_and_one_holdback(self):
        """
        Tests the PDFWriter writes correctly an accounting document with three
        items and one holdback.
        """
        self.acc_doc.acc_items = self._build_acc_items_list(3)
        self.acc_doc.acc_items[1].holdback_rate = 5.0
        self._run_pdf_test("%s_three_items_and_one_holdback.pdf"
                           % self.entity_name)

    def test_write_three_items_no_vat_and_one_holdback(self):
        """
        Tests the PDFWriter writes correctly an accounting document with three
        items and one holdback.
        """
        self.acc_doc.acc_items = self._build_acc_items_list(3)
        for i in range(3):
            self.acc_doc.acc_items[i].vat_rate = None
        self.acc_doc.acc_items[1].holdback_rate = 5.0
        self.acc_doc.acc_items[1].holdback_on_vat = True
        self._run_pdf_test("%s_three_items_no_vat_and_one_holdback.pdf"
                           % self.entity_name)

    def test_write_three_items_and_holdbacks(self):
        """
        Tests the PDFWriter writes correctly an accounting document with three
        items and one holdback.
        """
        self.acc_doc.acc_items = self._build_acc_items_list(3)
        for i in range(3):
            self.acc_doc.acc_items[i].holdback_rate = 5.0*(i+1)
            self.acc_doc.acc_items[i].holdback_on_vat = True
        self.acc_doc.acc_items[0].holdback_on_vat = False
        self.acc_doc.acc_items[2].vat_rate = None
        self._run_pdf_test("%s_three_items_and_holdbacks.pdf"
                           % self.entity_name)

    def test_write_three_items_and_remarks(self):
        """
        Tests the PDFWriter writes correctly an accounting document with three
        items and some preliminary remarks.
        """
        self.acc_doc.acc_items = self._build_acc_items_list(3)
        self.acc_doc.remarks.append("Remark 1 "*15)
        self.acc_doc.remarks.append("Remark 2 ")        
        self._run_pdf_test("%s_three_items_and_remarks.pdf" % self.entity_name)


    def test_write_various_items_total_on_page_2(self):
        """
        Tests the PDFWriter writes correctly an accounting document with
        various items and a total that can't be displayed on page 1 (so it's
        on page 2).
        """
        self.acc_doc.acc_items = self._build_acc_items_list(6)
        self._run_pdf_test("%s_various_items_total_on_page_2.pdf"
                           % self.entity_name)

    def test_write_various_items_on_two_pages(self):
        """
        Tests the PDFWriter writes correctly an accounting document with
        various items displayed on two pages (too much items for one single
        page).
        """
        self.acc_doc.acc_items = self._build_acc_items_list(10)
        self._run_pdf_test("%s_various_items_on_two_pages.pdf"
                           % self.entity_name)

    def test_write_one_item_and_logo(self):
        """
        Tests the PDFWriter writes correctly an accounting document with only
        one item and a logo for the company.
        """
        self.acc_doc.acc_items = self._build_acc_items_list(1)
        self.acc_doc.cfg_data.company_logo = unicode(self.datapath("mpe.png"))
        self._run_pdf_test("%s_one_item_and_logo.pdf" % self.entity_name)



# -*- coding: utf-8 -*-
"""
``pybill.lib.config`` package contains all the code that manages the
configurations.

The only public class that should be used outside of this package is the
:class:`~pybill.lib.config.register.ConfigRegister` class that is imported and
available in the namespace of this package.
"""
__docformat__ = "restructuredtext en"

# Imports the classes that must be used from the outside.
from pybill.lib.config.register import ConfigRegister



# -*- coding: utf-8 -*-
"""
``pybill.lib.pdfwriters.flowgenerators`` sub-package contains the definition of
the flow generation classes used in the PDF generation with ReportLab
`platypus`.

These classes generate the flow content of a bill, a claim form, a downpayment,
a pro-forma or a debit: title, various paragraphes, main table detailling the
items, and total table. This flow content will be displayed on various pages by
the :class:`~pybill.lib.pdfwriters.templates.doc_templates.AccDocDocumentTemplate``
document template thanks to the platypus machinery.

The sub-package is divided in several modules, each defining one of the flow
generator.

``abstract`` module
===================

.. automodule:: pybill.lib.pdfwriters.flowgenerators.abstract

``bill_generator`` module
=========================

.. automodule:: pybill.lib.pdfwriters.flowgenerators.bill_generator

``claimform_generator`` module
==============================

.. automodule:: pybill.lib.pdfwriters.flowgenerators.claimform_generator

``debit_generator`` module
==========================

.. automodule:: pybill.lib.pdfwriters.flowgenerators.debit_generator

``downpayment_generator`` module
================================

.. automodule:: pybill.lib.pdfwriters.flowgenerators.downpayment_generator

``proforma_generator`` module
=============================

.. automodule:: pybill.lib.pdfwriters.flowgenerators.proforma_generator
"""
__docformat__ = "restructuredtext en"

# Imports the classes exported by this package.

from pybill.lib.pdfwriters.flowgenerators.bill_generator import \
                                                              BillFlowGenerator
from pybill.lib.pdfwriters.flowgenerators.claimform_generator import \
                                                         ClaimFormFlowGenerator
from pybill.lib.pdfwriters.flowgenerators.downpayment_generator import \
                                                       DownpaymentFlowGenerator
from pybill.lib.pdfwriters.flowgenerators.debit_generator import \
                                                             DebitFlowGenerator
from pybill.lib.pdfwriters.flowgenerators.proforma_generator import \
                                                          ProFormaFlowGenerator



# -*- coding: utf-8 -*-

from logilab.common.testlib import TestCase, unittest_main
from StringIO import StringIO

from pybill.lib.config.entities import ConfigData

from pybill.lib.config.xmlwriters import save_config
from pybill.lib.errors import PyBillWritingException


EXPECTED = u"""<?xml version="1.0" encoding="UTF-8"?>

<pbc:config xmlns:pbc="http://www.logilab.org/2010/PyBillConfig" pbc:format-version="PBC-1.0" pbc:name="Z0">
  <pbc:company>
    <pbc:logo-file>A1</pbc:logo-file>
    <pbc:orgname>A2</pbc:orgname>
    <pbc:address>
      <pbc:street>A3</pbc:street>
      <pbc:street>A4</pbc:street>
      <pbc:pob>A5</pbc:pob>
      <pbc:postcode>A6</pbc:postcode>
      <pbc:city>A7</pbc:city>
      <pbc:state>A8</pbc:state>
      <pbc:country>A9</pbc:country>
      <pbc:phone>A10</pbc:phone>
      <pbc:fax>A11</pbc:fax>
      <pbc:web>A12</pbc:web>
      <pbc:email>A13</pbc:email>
    </pbc:address>
  </pbc:company>
  <pbc:bank-data>
    <pbc:line>B1</pbc:line>
    <pbc:line>B2</pbc:line>
  </pbc:bank-data>
  <pbc:footer>
    <pbc:line>D1</pbc:line>
    <pbc:line>D2</pbc:line>
  </pbc:footer>
  <pbc:agreement-intro>
    <pbc:line>C1</pbc:line>
    <pbc:line>C2</pbc:line>
  </pbc:agreement-intro>
  <pbc:number-separators>
    <pbc:sign>F1</pbc:sign>
    <pbc:thousands>F2</pbc:thousands>
    <pbc:digits>F3</pbc:digits>
  </pbc:number-separators>
  <pbc:localisation>
    <pbc:colon>E0</pbc:colon>
    <pbc:phone-kw>E1</pbc:phone-kw>
    <pbc:fax-kw>E2</pbc:fax-kw>
    <pbc:web-kw>E3</pbc:web-kw>
    <pbc:email-kw>E4</pbc:email-kw>
    <pbc:doc-ref-kw>E5</pbc:doc-ref-kw>
    <pbc:receiver-kw>E9</pbc:receiver-kw>
    <pbc:sender-kw>E10</pbc:sender-kw>
    <pbc:on-date>E11</pbc:on-date>
    <pbc:bill>E12</pbc:bill>
    <pbc:claim-form>E13</pbc:claim-form>
    <pbc:downpayment>E14</pbc:downpayment>
    <pbc:debit>E16</pbc:debit>
    <pbc:pro-forma>E15</pbc:pro-forma>
    <pbc:number>E17</pbc:number>
    <pbc:dated>E18</pbc:dated>
    <pbc:valid-until>E19</pbc:valid-until>
    <pbc:intro-detail>E20</pbc:intro-detail>
    <pbc:quantity>E21</pbc:quantity>
    <pbc:description>E22</pbc:description>
    <pbc:vat-rate>E23</pbc:vat-rate>
    <pbc:unit-price>E24</pbc:unit-price>
    <pbc:tf-unit-price>E25</pbc:tf-unit-price>
    <pbc:price>E26</pbc:price>
    <pbc:tf-price>E27</pbc:tf-price>
    <pbc:it-price>E28</pbc:it-price>
    <pbc:holdback-on>E29</pbc:holdback-on>
    <pbc:ita-est>E30</pbc:ita-est>
    <pbc:total>E31</pbc:total>
    <pbc:tf-total>E32</pbc:tf-total>
    <pbc:vat-amount>E33</pbc:vat-amount>
    <pbc:it-total>E34</pbc:it-total>
    <pbc:including>E35</pbc:including>
    <pbc:holdback>E36</pbc:holdback>
    <pbc:on-tf>E37</pbc:on-tf>
    <pbc:on-vat>E38</pbc:on-vat>
    <pbc:debit-total>E39</pbc:debit-total>
    <pbc:charged-downpayment>E40</pbc:charged-downpayment>
    <pbc:charged-on>E41</pbc:charged-on>
    <pbc:issued-debit>E42</pbc:issued-debit>
    <pbc:issued-on>E43</pbc:issued-on>
    <pbc:to-be-paid>E44</pbc:to-be-paid>
    <pbc:payment-terms>E45</pbc:payment-terms>
    <pbc:to-bring-forward>E46</pbc:to-bring-forward>
    <pbc:carry-forward>E47</pbc:carry-forward>
    <pbc:your-ref-kw>E101</pbc:your-ref-kw>
    <pbc:purch-ref-kw>E102</pbc:purch-ref-kw>
    <pbc:supplier-ref-kw>E103</pbc:supplier-ref-kw>
  </pbc:localisation>
</pbc:config>
"""

LOCAL_DICT = { u'colon'              : u"E0",
               u'phone-kw'           : u"E1",
               u'fax-kw'             : u"E2",
               u'web-kw'             : u"E3",
               u'email-kw'           : u"E4",
               u'doc-ref-kw'         : u"E5",
               u'receiver-kw'        : u"E9",
               u'sender-kw'          : u"E10",
               u'on-date'            : u"E11",
               u'bill'               : u"E12",
               u'claim-form'         : u"E13",
               u'downpayment'        : u"E14",
               u'pro-forma'          : u"E15",
               u'debit'              : u"E16",
               u'number'             : u"E17",
               u'dated'              : u"E18",
               u'valid-until'        : u"E19",
               u'intro-detail'       : u"E20",
               u'quantity'           : u"E21",
               u'description'        : u"E22",
               u'vat-rate'           : u"E23",
               u'unit-price'         : u"E24",
               u'tf-unit-price'      : u"E25",
               u'price'              : u"E26",
               u'tf-price'           : u"E27",
               u'it-price'           : u"E28",
               u'holdback-on'        : u"E29",
               u'ita-est'            : u"E30",
               u'total'              : u"E31",
               u'tf-total'           : u"E32",
               u'vat-amount'         : u"E33",
               u'it-total'           : u"E34",
               u'including'          : u"E35",
               u'holdback'           : u"E36",
               u'on-tf'              : u"E37",
               u'on-vat'             : u"E38",
               u'debit-total'        : u"E39",
               u'charged-downpayment': u"E40",
               u'charged-on'         : u"E41",
               u'issued-debit'       : u"E42",
               u'issued-on'          : u"E43",
               u'to-be-paid'         : u"E44",
               u'payment-terms'      : u"E45",
               u'to-bring-forward'   : u"E46",
               u'carry-forward'      : u"E47",
               u'your-ref-kw'        : u"E101",
               u'purch-ref-kw'       : u"E102",
               u'supplier-ref-kw'    : u"E103",
             }


class ConfigXMLWriterTest(TestCase):
    def setUp(self):
        """
        Called before each test of this class. Builds a config object.
        """
        self.cfg = ConfigData()
        self.cfg.reference = u"Test"
        self.cfg.name = u"Z0"
        self.cfg.company_logo = u"A1"
        self.cfg.company_address.name = u"A2"
        self.cfg.company_address.streets = [u"A3", u"A4"]
        self.cfg.company_address.post_box = u"A5"
        self.cfg.company_address.post_code = u"A6"
        self.cfg.company_address.city = u"A7"
        self.cfg.company_address.state = u"A8"
        self.cfg.company_address.country = u"A9"
        self.cfg.company_address.phone = u"A10"
        self.cfg.company_address.fax = u"A11"
        self.cfg.company_address.web = u"A12"
        self.cfg.company_address.email = u"A13"
        self.cfg.bank_data_lines = [u"B1", u"B2"]
        self.cfg.agreement_intro_lines = [u"C1", u"C2"]
        self.cfg.footer_lines = [u"D1", u"D2"]
        self.cfg.number_separators = {u"sign": u"F1",
                                      u"thousands": u"F2",
                                      u"digits": u"F3"}
        self.cfg.local_phrases = LOCAL_DICT

    def test_write_data(self):
        """
        Tests the XMLConfigReader writes correctly an XML config file
        """
        # uses the XML parser sympathy to pass it a string stream instead of a
        # filename
        xml_result = StringIO()
        save_config(self.cfg, xml_result)
        self.assertMultiLineEqual(xml_result.getvalue(), EXPECTED)


# definitions for automatic unit testing

if __name__ == '__main__':
    unittest_main()



# -*- coding: utf-8 -*-

from pybill.lib import ACCURACY, PBD_1_0
from pybill.lib.entities.accounting_docs import AccItem


class AbstractAccDocTest:
    """
    Base class factorizing the common behaviour for the tests of the various
    sort of bills.

    We should not define test methods on this object as it doesn't inherit from
    ``TestCase``... but it is so much simpler to do this and Python is a nice
    dynamic language that doesn't complain about this.

    :ivar acc_doc: Accounting object that will be tested. This object is
                   actually built by the child classes before each test in
                   the ``setUp`` method.
    :type acc_doc: `pybill.lib.entities.accounting_docs.GenericAccountingDoc`
    """
    def __init__(self):
        """
        Just defines the ``acc_doc`` attribute without setting it.
        """
        self.acc_doc = None

    def build_vat_example(self):
        """
        Creates an accounting document containing items with various VAT
        rates.
        """
        item = AccItem()
        item.quantity = 13.0
        item.unit_price = 14.0 
        self.acc_doc.acc_items.append(item)
        item = AccItem()
        item.quantity = 13.0
        item.unit_price = 17.0 
        item.vat_rate = 1.0
        self.acc_doc.acc_items.append(item)
        for i in range(3):
            item = AccItem()
            item.quantity = 5.0
            item.unit_price = 7.0 
            item.vat_rate = 1.0+(20.0-i*10.0)
            self.acc_doc.acc_items.append(item)

    def build_standard_example(self, with_holdbacks=False):
        """
        Creates an accounting document containing several items.

        :parameter with_holdbacks: When set to ``True``, holdbacks are
                                   inserted in the items.
        :type with_holdbacks: bool
        """
        item = AccItem()
        item.quantity = 17.0
        item.unit_price = 13.0
        if with_holdbacks:
            item.holdback_rate = 5.0
            item.holdback_on_vat = True
        self.acc_doc.acc_items.append(item)
        for i in range(3):
            item = AccItem()
            item.quantity = 5.0*(i+1)
            item.unit_price = 7.0*(i+1)
            item.vat_rate = 19.6
            if with_holdbacks and i in [1,2]:
                item.holdback_rate = 5.0
            if with_holdbacks and i in [0,1]:
                item.holdback_on_vat = True
            self.acc_doc.acc_items.append(item)

    def build_rounding_example(self):
        """
        Creates an accounting document containing several items. The totals
        of these items are rounded to ``ACCURACY`` digits.
        """
        for i in range(4):
            item = AccItem()
            item.quantity = 5.009
            item.unit_price = 7.0
            item.vat_rate = 19.6
            item.holdback_rate = 5.0
            item.holdback_on_vat = True
            self.acc_doc.acc_items.append(item)

    def test_empty_object(self):
        """
        Tests an empty object contains nothing and has totals of 0.0
        """
        self.assertEqual(self.acc_doc.cfg_data, None)
        self.assertEqual(self.acc_doc.account_numbers, {})
        self.assertEqual(self.acc_doc.origin_format, PBD_1_0)
        self.assertEqual(self.acc_doc.acc_items, [])
        self.assertEqual(self.acc_doc.get_all_vat_rates(), set())
        self.assertEqual(self.acc_doc.get_vat_amount_for_rate(19.6),
                          round(0.0, ACCURACY))
        self.assertEqual(self.acc_doc.get_price_total(),
                          round(0.0, ACCURACY))
        self.assertEqual(self.acc_doc.get_holdback_amount(),
                          round(0.0, ACCURACY))
        self.assertEqual(self.acc_doc.get_holdback_vat_amount(),
                          round(0.0, ACCURACY))
        self.assertEqual(self.acc_doc.get_including_taxes_total(),
                          round(0.0, ACCURACY))
        self.assertEqual(self.acc_doc.get_to_be_paid_holdback_free(),
                          round(0.0, ACCURACY))
        self.assertEqual(self.acc_doc.get_to_be_paid_holdback(),
                          round(0.0, ACCURACY))
        self.assertEqual(self.acc_doc.get_to_be_paid_vat(),
                          round(0.0, ACCURACY))
        self.assertEqual(self.acc_doc.get_to_be_paid(),
                          round(0.0, ACCURACY))

    def test_has_vat_rates_with_vat_rates(self):
        """
        Tests the method saying if the document contains VAT rates when VAT
        rates are defined.
        """
        self.build_standard_example()
        self.assertEqual(self.acc_doc.has_vat_rates(), True)

    def test_has_vat_rates_without_vat_rate(self):
        """
        Tests the method saying if the document contains VAT rates when no VAT
        rate is defined.
        """
        self.build_standard_example()
        for item in self.acc_doc.acc_items:
            item.vat_rate = None
        self.assertEqual(self.acc_doc.has_vat_rates(), False)
        
    def test_has_holdbacks_with_holdbacks(self):
        """
        Tests the method saying if the document contains holdbacks when
        holdback rates are defined on some items.
        """
        self.build_standard_example()
        self.acc_doc.acc_items[1].holdback_rate = 5.0
        self.assertEqual(self.acc_doc.has_holdbacks(), True)

    def test_has_holdbacks_without_holdback(self):
        """
        Tests the method saying if the document contains holdbacks when no
        holdback rate is defined.
        """
        self.build_standard_example()
        self.assertEqual(self.acc_doc.has_holdbacks(), False)

    def test_has_vat_holdbacks_with_vat_holdbacks(self):
        """
        Tests the method saying if the document contains vat holdbacks when
        holdback rates are defined on some items and some of these holdbacks
        are applied on VAT.
        """
        self.build_standard_example()
        self.acc_doc.acc_items[0].holdback_rate = 10.0
        self.acc_doc.acc_items[1].holdback_rate = 5.0
        self.acc_doc.acc_items[1].holdback_on_vat = True
        self.assertEqual(self.acc_doc.has_vat_holdbacks(), True)

    def test_has_vat_holdbacks_with_no_vat_holdbacks(self):
        """
        Tests the method saying if the document contains vat holdbacks when
        holdback rates are defined on some items but these holdbacks are not
        applied on VAT.
        """
        self.build_standard_example()
        self.acc_doc.acc_items[0].holdback_rate = 10.0
        self.acc_doc.acc_items[1].holdback_rate = 5.0
        self.acc_doc.acc_items[1].holdback_on_vat = False
        self.acc_doc.acc_items[2].holdback_on_vat = True
        self.assertEqual(self.acc_doc.has_vat_holdbacks(), False)

    def test_has_vat_holdbacks_without_vat_rates(self):
        """
        Tests the method saying if the document contains vat holdbacks when
        no VAT rates are defined.
        """
        self.build_standard_example()
        for item in self.acc_doc.acc_items:
            item.vat_rate = None
        self.acc_doc.acc_items[1].holdback_rate = 5.0
        self.acc_doc.acc_items[1].holdback_on_vat = True
        self.assertEqual(self.acc_doc.has_vat_holdbacks(), False)

    def test_has_vat_holdbacks_without_holdback(self):
        """
        Tests the method saying if the document contains holdbacks when no
        holdback rate is defined.
        """
        self.build_standard_example()
        self.assertEqual(self.acc_doc.has_vat_holdbacks(), False)

    def test_get_all_vat_rates(self):
        """
        Tests the method giving the different VAT rates used in the document.
        """
        self.build_vat_example()        
        self.assertEqual(self.acc_doc.get_all_vat_rates(),
                          set([1.0,11.0,21.0]))

    def test_get_no_vat_rate(self):
        """
        Tests the method giving the different VAT rates used in the document
        when no vat rate is defined.
        """
        self.build_standard_example()
        for item in self.acc_doc.acc_items:
            item.vat_rate = None
        self.assertEqual(self.acc_doc.get_all_vat_rates(), set())

    def test_get_vat_amount_for_rate(self):
        """
        Tests the method giving the VAT amount for a given VAT rate.
        """
        self.build_vat_example()
        self.assertEqual(self.acc_doc.get_vat_amount_for_rate(1.0),
                          round(2.56, ACCURACY))
        self.assertEqual(self.acc_doc.get_vat_amount_for_rate(11.0),
                          round(3.85, ACCURACY))
        self.assertEqual(self.acc_doc.get_vat_amount_for_rate(21.0),
                          round(7.35, ACCURACY))
        self.assertEqual(self.acc_doc.get_vat_amount_for_rate(19.6),
                          round(0.0, ACCURACY))

    def test_get_price_total(self):
        """
        Tests the computation of the tax free total.
        """
        self.build_standard_example()
        self.assertEqual(self.acc_doc.get_price_total(),
                          round(711.0, ACCURACY))

    def test_get_including_taxes_total(self):
        """
        Tests the computation of the including taxes total.
        """
        self.build_standard_example()
        self.assertEqual(self.acc_doc.get_including_taxes_total(),
                          round(807.04, ACCURACY))

    def test_get_holdback_amount(self):
        """
        Tests the computation of the amount of holdbacks.
        """
        self.build_standard_example()
        self.assertEqual(self.acc_doc.get_holdback_amount(),
                          round(0.0, ACCURACY))

    def test_get_holdback_vat_amount(self):
        """
        Tests the computation of the amount of holdbacks on vat.
        """
        self.build_standard_example()
        self.assertEqual(self.acc_doc.get_holdback_vat_amount(),
                          round(0.0, ACCURACY))

    def test_get_total_without_holdback_to_be_paid(self):
        """
        Tests the computation of the total without holdbacks to be paid.
        """
        self.build_standard_example()
        self.assertEqual(self.acc_doc.get_to_be_paid_holdback_free(),
                          round(807.04, ACCURACY))

    def test_get_total_holdback_to_be_paid(self):
        """
        Tests the computation of the total of holdbacks to be paid.
        """
        self.build_standard_example()
        self.assertEqual(self.acc_doc.get_to_be_paid_holdback(),
                          round(0.0, ACCURACY))

    def test_get_total_vat_to_be_paid(self):
        """
        Tests the computation of the total of VAT to be paid.
        """
        self.build_standard_example()
        self.assertEqual(self.acc_doc.get_to_be_paid_vat(),
                          round(96.04, ACCURACY))

    def test_get_total_to_be_paid(self):
        """
        Tests the computation of the total to be paid.
        """
        self.build_standard_example()
        self.assertEqual(self.acc_doc.get_to_be_paid(),
                          round(807.04, ACCURACY))

    def test_get_vat_amount_for_rate_with_holdback(self):
        """
        Tests the method giving the VAT amount for a given VAT rate when
        holdbacks are defined on some items.
        """
        self.build_vat_example()
        for i in range(len(self.acc_doc.acc_items)):
            item = self.acc_doc.acc_items[i]
            if i in [0,1,3]:
                item.holdback_rate = 5.0
            if i in [0,1,2]:
                item.holdback_on_vat = True
        self.assertEqual(self.acc_doc.get_vat_amount_for_rate(1.0),
                          round(2.56, ACCURACY))
        self.assertEqual(self.acc_doc.get_vat_amount_for_rate(11.0),
                          round(3.85, ACCURACY))
        self.assertEqual(self.acc_doc.get_vat_amount_for_rate(21.0),
                          round(7.35, ACCURACY))
        self.assertEqual(self.acc_doc.get_vat_amount_for_rate(19.6),
                          round(0.0, ACCURACY))

    def test_get_price_total_with_holdback(self):
        """
        Tests the computation of the tax free total when holdbacks are defined
        on some items.
        """
        self.build_standard_example(with_holdbacks=True)
        self.assertEqual(self.acc_doc.get_price_total(),
                          round(711.0, ACCURACY))

    def test_get_including_taxes_total_with_holdback(self):
        """
        Tests the computation of the including taxes total when holdbacks are
        defined on some items.
        """
        self.build_standard_example(with_holdbacks=True)
        self.assertEqual(self.acc_doc.get_including_taxes_total(),
                          round(807.04, ACCURACY))

    def test_get_holdback_amount_with_holdback(self):
        """
        Tests the computation of the amount of holdbacks when holdbacks are
        defined on some items.
        """
        self.build_standard_example(with_holdbacks=True)
        self.assertEqual(self.acc_doc.get_holdback_amount(),
                          round(33.80, ACCURACY))

    def test_get_holdback_vat_amount_with_holdback(self):
        """
        Tests the computation of the amount of holdbacks on vat when holdbacks
        are defined on some items.
        """
        self.build_standard_example(with_holdbacks=True)
        self.assertEqual(self.acc_doc.get_holdback_vat_amount(),
                          round(1.372, ACCURACY))

    def test_get_total_without_holdback_to_be_paid_with_holdback(self):
        """
        Tests the computation of the total without holdbacks to be paid when
        holdbacks are defined on some items.
        """
        self.build_standard_example(with_holdbacks=True)
        self.assertEqual(self.acc_doc.get_to_be_paid_holdback_free(),
                          round(771.868, ACCURACY))

    def test_get_total_holdback_to_be_paid_with_holdback(self):
        """
        Tests the computation of the total of holdbacks to be paid when
        holdbacks are defined on some items.
        """
        self.build_standard_example(with_holdbacks=True)
        self.assertEqual(self.acc_doc.get_to_be_paid_holdback(),
                          round(35.172, ACCURACY))

    def test_get_total_vat_to_be_paid_with_holdback(self):
        """
        Tests the computation of the total of VAT to be paid when holdbacks are
        defined on some items.
        """
        self.build_standard_example(with_holdbacks=True)
        self.assertEqual(self.acc_doc.get_to_be_paid_vat(),
                          round(96.04, ACCURACY))

    def test_get_total_to_be_paid_with_holdback(self):
        """
        Tests the computation of the total to be paid when holdbacks are
        defined on some items.
        """
        self.build_standard_example(with_holdbacks=True)
        self.assertEqual(self.acc_doc.get_to_be_paid(),
                          round(807.04, ACCURACY))

    def test_get_vat_amount_for_rate_when_rounding(self):
        """
        Tests the method giving the VAT amount for a given VAT rate when
        rounding to ``ACCURACY`` is done.
        """
        self.build_rounding_example()
        self.assertEqual(self.acc_doc.get_vat_amount_for_rate(19.6),
                          round(27.44, ACCURACY))

    def test_get_price_total_when_rounding(self):
        """
        Tests the computation of the tax free total when rounding to
        ``ACCURACY`` is done.
        """
        self.build_rounding_example()
        self.assertEqual(self.acc_doc.get_price_total(),
                          round(140.00, ACCURACY))

    def test_get_including_taxes_total_when_rounding(self):
        """
        Tests the computation of the including taxes total when rounding to
        ``ACCURACY`` is done.
        """
        self.build_rounding_example()
        self.assertEqual(self.acc_doc.get_including_taxes_total(),
                          round(167.44, ACCURACY))

    def test_get_holdback_amount_when_rounding(self):
        """
        Tests the computation of the amount of holdbacks when rounding to
        ``ACCURACY`` is done.
        """
        self.build_rounding_example()
        self.assertEqual(self.acc_doc.get_holdback_amount(),
                          round(7.0, ACCURACY))

    def test_get_holdback_vat_amount_when_rounding(self):
        """
        Tests the computation of the amount of holdbacks on vat when rounding
        to ``ACCURACY`` is done.
        """
        self.build_rounding_example()
        self.assertEqual(self.acc_doc.get_holdback_vat_amount(),
                          round(1.36, ACCURACY))

    def test_get_total_without_holdback_to_be_paid_when_rounding(self):
        """
        Tests the computation of the total without holdbacks to be paid when
        rounding to ``ACCURACY`` is done.
        """
        self.build_rounding_example()
        self.assertEqual(self.acc_doc.get_to_be_paid_holdback_free(),
                          round(159.08, ACCURACY))

    def test_get_total_holdback_to_be_paid_when_rounding(self):
        """
        Tests the computation of the total of holdbacks to be paid when
        rounding to ``ACCURACY`` is done.
        """
        self.build_rounding_example()
        self.assertEqual(self.acc_doc.get_to_be_paid_holdback(),
                          round(8.36, ACCURACY))

    def test_get_total_vat_to_be_paid_when_rounding(self):
        """
        Tests the computation of the total of VAT to be paid when rounding to
        ``ACCURACY`` is done.
        """
        self.build_rounding_example()
        self.assertEqual(self.acc_doc.get_to_be_paid_vat(),
                          round(27.44, ACCURACY))

    def test_get_total_to_be_paid_when_rounding(self):
        """
        Tests the computation of the total to be paid when rounding to
        ``ACCURACY`` is done.
        """
        self.build_rounding_example()
        self.assertEqual(self.acc_doc.get_to_be_paid(),
                          round(167.44, ACCURACY))



# -*- coding: utf-8 -*-

from logilab.common.testlib import TestCase, unittest_main

import datetime
from pybill.lib.errors import PyBillProcessingException
from pybill.lib.accounting.utils import Entry


class EntryTest(TestCase):
    def setUp(self):
        """
        Called before each test from this class
        """
        self.entry = Entry(datetime.date(2009, 11, 06), u"EXPLANATION")

    def test_empty_object(self):
        self.assertEqual(self.entry.date, datetime.date(2009, 11, 06))
        self.assertEqual(self.entry.explanation, u"EXPLANATION")
        self.assertEqual(self.entry.debits,  {})
        self.assertEqual(self.entry.credits, {})

    def test_add_credit_movement(self):
        self.entry.add_movement(u"44", 11.0)
        self.assertEqual(self.entry.debits,  {})
        self.assertEqual(self.entry.credits, {u"44": [11.0]})

    def test_add_debit_movement(self):
        self.entry.add_movement(u"44", -11.0)
        self.assertEqual(self.entry.credits, {})
        self.assertEqual(self.entry.debits,  {u"44": [11.0]})

    def test_add_credit_movement_when_rounding(self):
        self.entry.add_movement(u"44", 11.145)
        self.assertEqual(self.entry.debits,  {})
        self.assertEqual(self.entry.credits, {u"44": [11.15]})

    def test_add_debit_movement_when_rounding(self):
        self.entry.add_movement(u"44", -11.145)
        self.assertEqual(self.entry.credits, {})
        self.assertEqual(self.entry.debits,  {u"44": [11.15]})

    def test_add_credit_debit_movements_on_same_account(self):
        self.entry.add_movement(u"44", 11.0)
        self.entry.add_movement(u"44", -9.0)
        self.assertEqual(self.entry.debits,  {u"44": [ 9.0]})
        self.assertEqual(self.entry.credits, {u"44": [11.0]})

    def test_add_credit_movements_on_same_account(self):
        self.entry.add_movement(u"44", 11.0)
        self.entry.add_movement(u"44", 13.0)
        self.assertEqual(self.entry.debits,  {})
        self.assertEqual(self.entry.credits, {u"44": [11.0, 13.0]})

    def test_add_credit_movements_on_two_accounts(self):
        self.entry.add_movement(u"44", 11.0)
        self.entry.add_movement(u"45",  9.0)
        self.assertEqual(self.entry.debits,  {})
        self.assertEqual(self.entry.credits, {u"44": [11.0], u"45": [9.0]})

    def test_add_debit_movements_on_same_account(self):
        self.entry.add_movement(u"44", -11.0)
        self.entry.add_movement(u"44", -13.0)
        self.assertEqual(self.entry.credits, {})
        self.assertEqual(self.entry.debits,  {u"44": [11.0, 13.0]})

    def test_add_debit_movements_on_two_accounts(self):
        self.entry.add_movement(u"44", -11.0)
        self.entry.add_movement(u"45",  -9.0)
        self.assertEqual(self.entry.credits, {})
        self.assertEqual(self.entry.debits,  {u"44": [11.0], u"45": [9.0]})

    def test_check_correct_entry(self):
        self.entry.debits = {u"44": [12.0, 3.0], u"45": [4.0]}
        self.entry.credits = {u"46": [19.0]}
        self.assertTrue(self.entry.check())

    def test_check_incorrect_entry(self):
        self.entry.debits = {u"44": [12.0, 1.0], u"45": [4.0]}
        self.entry.credits = {u"46": [19.0], u"47": [1.0]}
        self.assertRaises(PyBillProcessingException, self.entry.check)


# definitions for automatic unit testing

if __name__ == '__main__':
    unittest_main()



# -*- coding: utf-8 -*-

from logilab.common.testlib import TestCase, unittest_main
from StringIO import StringIO

from pybill.lib.config.entities import ConfigData
from pybill.lib.entities.accounting_docs import Bill, AccItem
from pybill.lib.entities.addresses import PersonAddress, OrganisationAddress

from pybill.lib.xmlwriters import write_accounting_doc
from pybill.lib.errors import PyBillWritingException


EXPECTED_XML_1_0 = u"""<?xml version="1.0" encoding="%s"?>
%s
<pbd:accounting-document xmlns:pbd="http://www.logilab.org/2010/PyBillDocument" pbd:format-version="PBD-1.0" pbd:type="bill">
  <pbd:metadata>
    <pbd:id>A1</pbd:id>
    <pbd:doc-ref>A2</pbd:doc-ref>
    <pbd:place>A3</pbd:place>
    <pbd:date>A4</pbd:date>
  </pbd:metadata>
  <pbd:address pbd:role="from">
    <pbd:firstname>B1</pbd:firstname>
    <pbd:surname>B2</pbd:surname>
    <pbd:phone>B3</pbd:phone>
    <pbd:email>B4</pbd:email>
  </pbd:address>
  <pbd:address pbd:role="to">
    <pbd:firstname>C1</pbd:firstname>
    <pbd:surname>C2</pbd:surname>
    <pbd:affiliation>
      <pbd:orgname>C3</pbd:orgname>
      <pbd:address>
        <pbd:street>C4</pbd:street>
        <pbd:postcode>C5</pbd:postcode>
        <pbd:city>C6</pbd:city>
      </pbd:address>
    </pbd:affiliation>
  </pbd:address>
  <pbd:items-list>
    <pbd:item>
      <pbd:qty pbd:digits="0">41.0</pbd:qty>
      <pbd:description>
        <pbd:title>D2</pbd:title>
        <pbd:detail>D3</pbd:detail>
      </pbd:description>
      <pbd:unit-price pbd:digits="2">44.0</pbd:unit-price>
      <pbd:vat-rate>45.0</pbd:vat-rate>
    </pbd:item>
    <pbd:item>
      <pbd:qty pbd:digits="0">51.0</pbd:qty>
      <pbd:description>
        <pbd:title>E2</pbd:title>
      </pbd:description>
      <pbd:unit-price pbd:digits="2">53.0</pbd:unit-price>
      <pbd:vat-rate>54.0</pbd:vat-rate>
    </pbd:item>
  </pbd:items-list>
  <pbd:payment-terms>F1</pbd:payment-terms>
</pbd:accounting-document>
"""


class AccDocXMLWriterTest(TestCase):
    def setUp(self):
        """
        Called before each test from this class. Builds the bill used in the
        tests.
        """
        self.accdoc = Bill(u"A1")
        self.accdoc.doc_ref = u"A2"
        self.accdoc.place = u"A3"
        self.accdoc.date = u"A4"
        self.accdoc.sender = PersonAddress()
        self.accdoc.sender.firstname = u"B1"
        self.accdoc.sender.surname = u"B2"
        self.accdoc.sender.phone = u"B3"
        self.accdoc.sender.email = u"B4"
        self.accdoc.receiver = PersonAddress()
        self.accdoc.receiver.firstname = u"C1"
        self.accdoc.receiver.surname = u"C2"
        self.accdoc.receiver.organisation = OrganisationAddress()
        self.accdoc.receiver.organisation.name = u"C3"
        self.accdoc.receiver.organisation.streets = [u"C4"]
        self.accdoc.receiver.organisation.post_code = u"C5"
        self.accdoc.receiver.organisation.city = u"C6"
        it1 = AccItem()
        it1.quantity = 41.0
        it1.title = u"D2"
        it1.details = [u"D3"]
        it1.unit_price = 44.0
        it1.vat_rate = 45.0
        self.accdoc.acc_items.append(it1)
        it2 = AccItem()
        it2.quantity = 51.0
        it2.title = u"E2"
        it2.unit_price = 53.0
        it2.vat_rate = 54.0
        self.accdoc.acc_items.append(it2)
        self.accdoc.payment_terms = u"F1"

    def test_write_doc_with_no_config(self):
        """
        Tests the writing of an accounting document that is not linked to
        a configuration object (even if this should never exist...)
        """
        xml_result = StringIO()
        # uses the writer sympathy to pass it a string stream instead of a
        # filename
        write_accounting_doc(self.accdoc, xml_result)
        self.assertMultiLineEqual(xml_result.getvalue(),
                              EXPECTED_XML_1_0 % (u"UTF-8", u""))

    def test_write_doc_with_config(self):
        """
        Tests the writing of an accounting document that is linked to
        a configuration object.
        """
        self.accdoc.cfg_data = ConfigData()
        self.accdoc.cfg_data.name = u"Test"
        xml_result = StringIO()
        # uses the writer sympathy to pass it a string stream instead of a
        # filename
        write_accounting_doc(self.accdoc, xml_result)
        self.assertMultiLineEqual(xml_result.getvalue(),
                EXPECTED_XML_1_0 % (u"UTF-8", u'\n<?pybill config="Test"?>\n'))

    def test_write_doc_with_config_and_encoding(self):
        """
        Tests the writing of an accounting document that is linked to
        a configuration object and that must be written in a different 
        encoding.
        """
        self.accdoc.cfg_data = ConfigData()
        self.accdoc.cfg_data.name = u"Test"
        xml_result = StringIO()
        # uses the writer sympathy to pass it a string stream instead of a
        # filename
        write_accounting_doc(self.accdoc, xml_result, encoding="ISO-8859-1")
        self.assertMultiLineEqual(xml_result.getvalue(),
           EXPECTED_XML_1_0 % (u"ISO-8859-1", u'\n<?pybill config="Test"?>\n'))

    def test_write_unknown_format(self):
        """
        Tests the writer doesn't write un unknown formats.
        """
        xml_result = StringIO()
        # uses the XML parser sympathy to pass it a string stream instead of a
        # filename
        self.assertRaises(PyBillWritingException, write_accounting_doc,
                          self.accdoc, xml_result, u"PDB-1.1")


# definitions for automatic unit testing

if __name__ == '__main__':
    unittest_main()



# -*- coding: utf-8 -*-
"""
PyBill is a software that can transform an XML description of accounting
documents (such as bills, claim forms, debits or downpayments) into PDF
printable documents.

The XML documents are defined into the `PyBill Document 1.0` format.
Configuration files also described in XML allow to have different renderings
(logo, sender address and internationalization). The chosen configuration can be
set into a processing instruction in the XML accounting document.

PyBill uses :mod:`lxml` library to read and write the XML files, and
:mod:`reportlab` library to produce the PDFs.
"""
__docformat__ = "restructuredtext en"



# -*- coding: utf-8 -*-
"""
``pybill.lib.controller`` module defines the main controller of PyBill.

This controller knows how to read the XML accounting documents and how to
produce PDF, using the other classes that exist in the ``pybill.lib`` module.

.. autoclass:: PyBillController
   :members:
   :undoc-members:
   :show-inheritance:
"""
__docformat__ = "restructuredtext en"

import os
import os.path as osp

from pybill.lib import PBD_1_0

from pybill.lib.config import ConfigRegister
from pybill.lib.xmlreaders import AccDocXMLReader
from pybill.lib.pdfwriters import PDFWriter
from pybill.lib.xmlwriters import write_accounting_doc
from pybill.lib.accounting.generator import EntriesGenerator


class PyBillController:
    """
    Main controller that implements all the operations available in PyBill
    software.

    This controller uses dedicated object for reading XML documents and
    producing PDF output.

    .. attribute:: config_register

        Attribute that contains a config register object. This object has got a
        library of config files read in the various config directories and can
        give the config data used by a given accounting entity.

        type: :class:`~pybill.lib.config.register.ConfigRegister`

    .. attribute:: accdoc_reader

        Attribute that contains an XML reader object. This object can read an
        XML file containing an accounting document, build the corresponding
        entity and give the name of the config file used by this entity.

        type: :class:`~pybill.lib.xmlreaders.accdoc.AccDocXMLReader`

    .. attribute:: pdf_writer

        Attribute that contains a PDF writer object. This object can produce a
        PDF file rendering an accounting entity.
        
        type: :class:`~pybill.lib.pdfwriters.writers.PDFWriter`

    .. attribute:: entries_generator

        Attribute containing an entries generator object. This object can export
        accounting entries corresponding to accounting entities into an XML
        file. The XML file is compatible with `pycompta` software.

        type: :class:`~pybill.lib.accounting.generator.EntriesGenerator`

    .. automethod:: __init__
    
    """
    def __init__(self):
        """
        Initializes a new ``PyBillController`` object.

        Builds the specialized objects that will be used during the operations.
        """
        self.config_register = ConfigRegister()
        self.accdoc_reader = AccDocXMLReader(self.config_register)
        self.pdf_writer = PDFWriter()
        self.entries_generator = EntriesGenerator()

    def build_entities_from_files(self, filenames, cfg_filename=None):
        """
        Builds the accounting entities from files containing XML data in
        PyBill Document formats (e.g. `PBD-1.0`).

        The names of the files are given in ``filenames`` list. The method
        returns a list of couples (accounting entity, base of filename where
        the entity was read).

        :parameter filenames: List of the names (strings) of the files that
                              contain accounting documents in XML PyBill
                              formats (e.g. `PBD-1.0`). These documents will
                              be read and stored as accounting entities
                              (``Bill`` objects, etc.)
        :type filenames: list of :class:`str`

        :parameter cfg_filename: Name of a file containing a configuration in
                                 XML format. When specified, this configuration
                                 will be used for all the accouting entities
                                 read from ``filenames`` files and will preempt
                                 any configuration specified in these files.
        :type cfg_filename: :class:`str`

        :returns: List of couples (accounting entity, base of filename where
                  this entity was read from). The accounting entity was built
                  from one of the file in ``filenames``. The base of filename
                  is the name of this file without its extension.
        :rtype:
            list of tuples (:class:`~pybill.lib.entities.accounting_docs.abstract.GenericAccountingDoc`, :class:`str`)
        """
        entities = []
        if cfg_filename is not None:
            cfg_filename = osp.abspath(cfg_filename)
        for filename in filenames:
            in_filename = osp.abspath(filename)
            accdoc = self.accdoc_reader.load_data(in_filename, cfg_filename)
            entities.append((accdoc, osp.splitext(in_filename)[0]))
        return entities

    def save_entities_to_files(self, entities, xml_encoding):
        """
        Saves the accounting entities in some XML files in latest PyBill
        format (\ `PBD-1.0`\ ).

        Actually, the ``entities`` list consists in a list of couples
        (accounting entity, name of the file where to save the XML document).

        If the file where the XML will be saved already exists, it is copied
        into a `.old` file before the XML is saved.

        :parameter entities: List of couples composed of an accounting entity
                             (e.g. ``Bill`` object) and the name of the file
                             where the XML representation of this entity will
                             be saved in.
        :type entities:
            list of tuples (:class:`~pybill.lib.entities.accounting_docs.abstract.GenericAccountingDoc`, :class:`str`)

        :parameter xml_encoding: Encoding used for writing the XML files
                                 describing the accounting entities.
        :type xml_encoding: :class:`str`
        """
        for accdoc, xml_filename in entities:
            if osp.exists(xml_filename):
                old_filename = "%s%sold" % (xml_filename, osp.extsep)
                if osp.exists(old_filename):
                    os.remove(old_filename)
                os.rename(xml_filename, old_filename)
            write_accounting_doc(accdoc, xml_filename, 
                                 format=PBD_1_0, encoding=xml_encoding)

    def write_pdf_to_files(self, entities):
        """
        Builds the PDF corresponding to accounting entities and saves them in
        files.

        Actually, the ``entities`` list consists in a list of couples
        (accounting entity, name of the file where to save the result of
        the PDF generation).

        :parameter entities: List of couples composed of an accounting entity
                             (e.g. ``Bill`` object) and the name of the file
                             where the PDF produced for this entity will be
                             saved in.
        :type entities:
            list of tuples (:class:`~pybill.lib.entities.accounting_docs.abstract.GenericAccountingDoc`, :class:`str`)
        """
        for accdoc, pdf_filename in entities:
            self.pdf_writer.write(accdoc, pdf_filename)

    def export_accounting_entries_to_xml_file(self, entities, 
                                              filename, xml_encoding):
        """
        Builds accounting entries from the accounting entities and save them
        in an XML file.

        For this method, the ``entities`` list only consists in a list of
        accounting entities.

        :parameter entities: List of accounting entities (e.g. ``Bill``
                             object).
        :type entities:
            list of :class:`~pybill.lib.entities.accounting_docs.abstract.GenericAccountingDoc`

        :parameter filename: Name of the file where the XML accounting entries
                             will be saved in.
        :type filename: :class:`str`

        :parameter xml_encoding: Encoding used for writing the XML file
                                 containing the accounting entries.
        :type xml_encoding: :class:`str`
        """
        self.entries_generator.reset_entries()
        for accdoc in entities:
            self.entries_generator.generate_entry(accdoc)
        self.entries_generator.write_entries_in_xml(filename, 
                                                    encoding=xml_encoding)



# -*- coding: utf-8 -*-

from logilab.common.testlib import TestCase, unittest_main
from os import path as osp

from pybill.lib.config.register import ConfigRegister


TEST_DATA_DIR = unicode(osp.join(osp.dirname(osp.abspath(__file__)), u"data"))


class TestableConfigRegister(ConfigRegister):
    """
    Class only defined for test purposes. Redefines the class
    constants in order to be able to do the tests.
    """
    SYSTEM_CONFIG_DIR = osp.join(TEST_DATA_DIR, u"system_dir")
    USER_CONFIG_DIR = osp.join(TEST_DATA_DIR, u"user_dir")
    DEFAULT_CONFIG_REF = u"default" 

class TestableConfigRegister2(ConfigRegister):
    """
    Class only defined for test purposes. Redefines the class
    constants in order to be able to do the tests.
    """
    SYSTEM_CONFIG_DIR = osp.join(TEST_DATA_DIR, u"system_dir")
    USER_CONFIG_DIR = osp.join(TEST_DATA_DIR, u"user_dir")
    DEFAULT_CONFIG_REF = u"non-existing-default" 


class ConfigRegisterTest(TestCase) :
    def setUp(self) :
        """
        Called before each test from this class
        """
        self.register = TestableConfigRegister()

    def test_real_class_constants(self):
        """
        Tests the value of the class constants in the real register class.
        """
        real_reg = ConfigRegister()
        self.assertEqual(real_reg.__class__.SYSTEM_CONFIG_DIR,
                          u"/etc/pybill/config/")
        self.assertEqual(real_reg.__class__.USER_CONFIG_DIR,
                          u"~/.config/pybill/")
        self.assertEqual(real_reg.__class__.DEFAULT_CONFIG_REF,
                          u"default")

    def test_find_file_defined_in_first_dir(self):
        """
        Tests the find_config_file method for a file that only exists in the
        user directory.
        """
        filename = self.register.find_config_file(u"config2")
        self.assertEqual(filename,
                          unicode(self.datapath(u"user_dir/config2.xml")))
        
    def test_find_file_defined_in_second_dir(self):
        """
        Tests the find_config_file method for a file that only exists in the
        system directory.
        """
        filename = self.register.find_config_file(u"config1")
        self.assertEqual(filename,
                          unicode(self.datapath(u"system_dir/config1.xml")))
        
    def test_find_file_defined_in_both_dirs(self):
        """
        Tests the find_config_file method for a file that exists in the user
        directory and the system directory.
        """
        filename = self.register.find_config_file(u"default")
        self.assertEqual(filename,
                          unicode(self.datapath(u"user_dir/default.xml")))

    def test_find_default_file(self):
        """
        Tests the find_config_file method for the file that contains the
        default configuration.
        """
        filename = self.register.find_config_file(
                                   self.register.__class__.DEFAULT_CONFIG_REF)
        self.assertEqual(filename,
                          unicode(self.datapath(u"user_dir/default.xml")))

    def test_find_file_from_filepath(self):
        """
        Tests the find_config_file method for a file defined by its file path.
        """
        filename = self.register.find_config_file(
                                unicode(self.datapath("user_dir/config2.xml")))
        self.assertEqual(filename,
                          unicode(self.datapath(u"user_dir/config2.xml")))

    def test_store_config(self):
        """
        Tests the store_config method.
        """
        self.register.store_config(u"DEFAULT2",
                             unicode(self.datapath(u"system_dir/default.xml")))
        self.assertTrue(u"DEFAULT2" in self.register.configs)
        cfg = self.register.configs[u"DEFAULT2"]
        self.assertEqual(cfg.reference, u"DEFAULT2")
        self.assertEqual(cfg.company_logo, u"B1")
        self.assertEqual(cfg.company_address.name, u"B2")
        self.assertEqual(cfg.local_phrases[u"bill"], u"B3")

    def test_store_config_from_filepath(self):
        """
        Tests the store_config method with a configuration defined by its
        filepath.
        """
        filepath = unicode(self.datapath(u"system_dir/config1.xml"))
        self.register.store_config(filepath, filepath)
        self.assertTrue(filepath in self.register.configs)
        cfg = self.register.configs[filepath]
        self.assertEqual(cfg.reference, filepath)
        self.assertEqual(cfg.company_logo, u"C1")
        self.assertEqual(cfg.company_address.name, u"C2")
        self.assertEqual(cfg.local_phrases[u"bill"], u"C3")

    def test_initialisation(self):
        """
        Tests the register was correctly initialized and has read the default
        configuration.
        """
        self.assertEqual(self.register.configs.keys(), [u"default"])
        cfg = self.register.configs[u"default"]
        self.assertEqual(cfg.reference, u"default")
        self.assertEqual(cfg.company_logo, u"A1")
        self.assertEqual(cfg.company_address.name, u"A2")
        self.assertEqual(cfg.local_phrases[u"bill"], u"A3")

    def test_initialisation_with_non_existing_default(self):
        """
        Tests the register is correctly initialized when the default
        configuration doesn't exist (the default configuration is then an
        empty configuration object).
        """
        other_reg = TestableConfigRegister2()
        self.assertEqual(other_reg.configs.keys(), [u"non-existing-default"])
        cfg = other_reg.configs[u"non-existing-default"]
        self.assertEqual(cfg.reference, None)
        self.assertEqual(cfg.company_logo, u"")
        self.assertEqual(cfg.company_address.name, u"")
        self.assertEqual(cfg.local_phrases, {})

    def test_get_read_config(self):
        """
        Tests the get_config method for an already read configuration.
        """
        self.register.store_config(u"config1",
                             unicode(self.datapath(u"system_dir/config1.xml")))
        cfg = self.register.get_config(u"config1")
        self.assertEqual(cfg.reference, u"config1")
        self.assertEqual(cfg.company_logo, u"C1")
        self.assertEqual(cfg.company_address.name, u"C2")
        self.assertEqual(cfg.local_phrases[u"bill"], u"C3")

    def test_get_non_read_config(self):
        """
        Tests the get_config method for a configuration that wasn't previously
        read.
        """
        self.register.store_config(u"config1",
                             unicode(self.datapath(u"system_dir/config1.xml")))
        cfg = self.register.get_config(u"config2")
        self.assertEqual(cfg.reference, u"config2")
        self.assertEqual(cfg.company_logo, u"D1")
        self.assertEqual(cfg.company_address.name, u"D2")
        self.assertEqual(cfg.local_phrases[u"bill"], u"D3")
        # Checks the newly-read configuration has been added in the register
        self.assertTrue(u"config2" in self.register.configs)
        cfg = self.register.configs[u"config2"]
        self.assertEqual(cfg.reference, u"config2")
        self.assertEqual(cfg.company_logo, u"D1")
        self.assertEqual(cfg.company_address.name, u"D2")
        self.assertEqual(cfg.local_phrases[u"bill"], u"D3")

    def test_get_non_existing_config(self):
        """
        Tests the get_config method for a configuration that doesn't exist (the
        method returns the default configuration).
        """
        self.register.store_config(u"config1",
                             unicode(self.datapath(u"system_dir/config1.xml")))
        cfg = self.register.get_config(u"config3")
        self.assertEqual(cfg.reference, u"default")
        self.assertEqual(cfg.company_logo, u"A1")
        self.assertEqual(cfg.company_address.name, u"A2")
        self.assertEqual(cfg.local_phrases[u"bill"], u"A3")
        # Checks the non existing configuration wasn't added in the register
        self.assertFalse(u"config3" in self.register.configs)

    def test_get_read_config_doesn_t_read_from_disk(self):
        """
        Tests the get_config method for an already read configuration doesn't
        read the configuration from the disk
        """
        self.register.store_config(u"config1",
                             unicode(self.datapath(u"system_dir/config1.xml")))
        self.register.configs[u"default"].company_logo = u"F45"
        cfg = self.register.get_config(u"default")
        self.assertEqual(cfg.reference, u"default")
        self.assertEqual(cfg.company_logo, u"F45")
        self.assertEqual(cfg.company_address.name, u"A2")
        self.assertEqual(cfg.local_phrases[u"bill"], u"A3")

    def test_get_config_from_filepath(self):
        """
        Tests the get_config method for a configuration defined by its filepath.
        """
        filepath = unicode(self.datapath(u"system_dir/config1.xml"))
        cfg = self.register.get_config(filepath)
        self.assertEqual(cfg.reference, filepath)
        self.assertEqual(cfg.company_logo, u"C1")
        self.assertEqual(cfg.company_address.name, u"C2")
        self.assertEqual(cfg.local_phrases[u"bill"], u"C3")
        self.assertTrue(filepath in self.register.configs)


# definitions for automatic unit testing

if __name__ == '__main__':
    unittest_main()



# -*- coding: utf-8 -*-

from logilab.common.testlib import TestCase, unittest_main

from pybill.lib.pdfwriters.utils import escape_text, format_number
from pybill.lib.pdfwriters.utils import TotalizableTable, NumberParagraph
from pybill.lib.pdfwriters import style_normal_left
from reportlab.platypus import Paragraph


class EscapeTextFunctionTest(TestCase):
    def test_no_escaping_characters(self):
        """
        Tests the ``escape_text`` function when the text contains no character
        to be escaped.
        """
        text = "   coucou c'est   moi !    "
        expected_text = "   coucou c'est   moi !    "
        self.assertEqual(expected_text, escape_text(text))

    def test_escaping_characters(self):
        """
        Tests the ``escape_text`` function when the text contains characters
        to be escaped.
        """
        text = "   coucou c'est   moi  & hop <>  !    "
        expected_text = "   coucou c'est   moi  &amp; hop &lt;&gt;  !    "
        self.assertEqual(expected_text, escape_text(text))

    def test_no_escaping_characters_and_normalizing(self):
        """
        Tests the ``escape_text`` function when the text contains no
        character to be escaped and when the normalizing feature is set.
        """
        text = "   coucou c'est   moi !    "
        expected_text = "coucou c'est moi !"
        self.assertEqual(expected_text, escape_text(text, True))

    def test_escaping_characters_and_normalizing(self):
        """
        Tests the ``escape_text`` function when the text contains characters
        to be escaped and when the normalizing feature is set.
        """
        text = "   coucou c'est   moi  & hop <>  !    "
        expected_text = "coucou c'est moi &amp; hop &lt;&gt; !"
        self.assertEqual(expected_text, escape_text(text, True))


class FormatNumberFunctionTest(TestCase):
    def setUp(self):
        """
        Called before each test.
        """
        self.seps = {u'sign': u' ',
                     u'thousands': u' ',
                     u'digits': u','}
        
    def test_number_is_None(self):
        """
        Tests the ``format_number`` function with a number that is None.
        """
        num = None
        expected_text = u""
        self.assertEqual(expected_text, format_number(num, 3, self.seps))

    def test_integer_less_than_thousand(self):
        """
        Tests the ``format_number`` function with an integer less than one
        thousand.
        """
        num = 532
        expected_text = u"532"
        self.assertEqual(expected_text, format_number(num, 0, self.seps))

    def test_integer_less_than_million(self):
        """
        Tests the ``format_number`` function with an integer between one
        thousand and one million.
        """
        num = 47032
        expected_text = u"47 032"
        self.assertEqual(expected_text, format_number(num, 0, self.seps))

    def test_great_integer(self):
        """
        Tests the ``format_number`` function with an integer greater than one
        million.
        """
        num = 2045008
        expected_text = u"2 045 008"
        self.assertEqual(expected_text, format_number(num, 0, self.seps))

    def test_integer_with_digits(self):
        """
        Tests the ``format_number`` function with an integer less than one
        thousand when asking for two digits.
        """
        num = 12345678
        expected_text = u"12 345 678,00"
        self.assertEqual(expected_text, format_number(num,
                                                       separators=self.seps))

    def test_float_rounding_to_lesser(self):
        """
        Tests the ``format_number`` function with a float, especially the
        rounding feature with two digits (case when rounding to lesser).
        """
        num = 532.8845512
        expected_text = u"532,88"
        self.assertEqual(expected_text, format_number(num,
                                                       separators=self.seps))

    def test_float_rounding_middle_to_greater(self):
        """
        Tests the ``format_number`` function with a float, especially the
        rounding feature with two digits (middle case that is rounded to
        greater).
        """
        num = 532.8855512
        expected_text = u"532,89"
        self.assertEqual(expected_text, format_number(num,
                                                       separators=self.seps))

    def test_float_rounding_to_greater(self):
        """
        Tests the ``format_number`` function with a float, especially the
        rounding feature with two digits (case when rounding to greater).
        """
        num = 532.8865512
        expected_text = u"532,89"
        self.assertEqual(expected_text, format_number(num,
                                                       separators=self.seps))

    def test_float_rounding_to_upper_integer(self):
        """
        Tests the ``format_number`` function with a float, especially the
        rounding feature with zero digits (case when rounding to upper integer).
        """
        num = 532.501
        expected_text = u"533"
        self.assertEqual(expected_text, format_number(num, 0, self.seps))

    def test_float_rounding_to_lower_integer(self):
        """
        Tests the ``format_number`` function with a float, especially the
        rounding feature with zero digits (case when rounding to lower integer).
        """
        num = 532.499
        expected_text = u"532"
        self.assertEqual(expected_text, format_number(num, 0, self.seps))

    def test_float_number_of_digits(self):
        """
        Tests the ``format_number`` function with a float, especially the
        number of digits.
        """
        num = 532.88555512
        expected_text = u"532,88556"
        self.assertEqual(expected_text, format_number(num, 5, self.seps))

    def test_great_float(self):
        """
        Tests the ``format_number`` function with an integer greater than one
        million, when four digits are asked
        """
        num = 12345678.123456
        expected_text = u"12 345 678,1235"
        self.assertEqual(expected_text, format_number(num, 4, self.seps))

    def test_negative_number(self):
        """
        Tests the ``format_number`` function with a negative number (float
        number by the way).
        """
        num = -12345678.123456
        expected_text = u"- 12 345 678,1235"
        self.assertEqual(expected_text, format_number(num, 4, self.seps))

    def test_default_separators(self):
        """
        Tests the ``format_number`` function with a negative float number
        when using the default separators for the formatting.
        """
        num = -12345678.123456
        expected_text = u"-12345678.1235"
        self.assertEqual(expected_text, format_number(num, 4))


class TotalizableTableTest(TestCase):
    def setUp(self):
        """
        Called before each test from this class
        """
        table_content = []
        # line 1
        line_content = []
        line_content.append(Paragraph(u"text1.1", style_normal_left))
        line_content.append(Paragraph(u"1", style_normal_left))
        line_content.append(NumberParagraph(1, u"1", style_normal_left))
        table_content.append(line_content)
        # line 2
        line_content = []
        line_content.append(Paragraph(u"text2.1", style_normal_left))
        line_content.append(Paragraph(u"20", style_normal_left))
        line_content.append(NumberParagraph(20.2, u"20", style_normal_left))
        table_content.append(line_content)
        # line 3
        line_content = []
        line_content.append(Paragraph(u"text3.1", style_normal_left))
        line_content.append(Paragraph(u"300", style_normal_left))
        line_content.append(NumberParagraph(300, u"300", style_normal_left))
        table_content.append(line_content)
        # Builds the table
        self.tot_table = TotalizableTable(table_content, [10,10,10], None)
    
    def test_total_computation(self):
        """
        Tests the computation of the total inside a ``TotalizableTable``.
        """
        expected_tot = 321.2
        computed_tot = self.tot_table.compute_total()
        self.assertEqual(expected_tot, computed_tot)


# definitions for automatic unit testing

if __name__ == '__main__':
    unittest_main()



# -*- coding: utf-8 -*-
"""
``pybill.lib.entities`` package defines the PyBill entities used all over
PyBill.

These entities represent the bills, the claim forms, the debits, the
downpayments, the pro-formas and all their inner objects.
"""
__docformat__ = "restructuredtext en"



# -*- coding: utf-8 -*-

from logilab.common.testlib import TestCase, unittest_main

from pybill.lib.entities.addresses import PersonAddress, OrganisationAddress


class PersonAddressTest(TestCase):
    def setUp(self):
        """
        Called before each test of this class.
        """
        self.addr = PersonAddress()
        self.addr.honorific  =   u" Mr               "
        self.addr.firstname  =   u" Robert           "
        self.addr.other_name =   u" T.               "
        self.addr.surname    =   u" Smith            "
        self.addr.lineage    =   u" Jr               "
        self.addr.streets    = [ u"                  ",
                                 u" Main street, 1st ",
                                 u" Yellow House     " ]
        self.addr.post_box   =   u" Post box    # 3  "
        self.addr.city       =   u" LONDON           "
        self.addr.post_code  =   u" 65789            "
        self.addr.state      =   u" (London County)  "
        self.addr.country    =   u" UNITED KINGDOM   "
        
    def test_get_postal_address_lines(self):
        """
        Test the method giving the lines of the postal address.
        """
        addr_lines = self.addr.get_postal_address_lines()
        expected_lines = [ u"Main street, 1st",
                           u"Yellow House",
                           u"Post box # 3",
                           u"65789 LONDON (London County)",
                           u"UNITED KINGDOM" ]
        self.assertEqual(addr_lines, expected_lines)

    def test_get_person_name(self) :
        name = self.addr.get_person_name()
        expected_name = u"Mr Robert T. Smith Jr"
        self.assertEqual(name, expected_name)


class OrganisationAddressTest(TestCase):
    def setUp(self):
        """
        Called before each test of this class.
        """
        self.addr = OrganisationAddress()
        self.addr.name       =   u" Stuff Corporation "
        self.addr.streets    = [ u"                   ",
                                 u" Main street, 1st  ",
                                 u" Yellow House      " ]
        self.addr.post_box   =   u" Post box    # 3   "
        self.addr.city       =   u" LONDON            "
        self.addr.post_code  =   u" 65789             "
        self.addr.state      =   u" (London County)   "
        self.addr.country    =   u" UNITED KINGDOM    "
        
    def test_get_postal_address_lines(self):
        """
        Test the method giving the lines of the postal address
        """
        addr_lines = self.addr.get_postal_address_lines()
        expected_lines = [ u"Main street, 1st",
                           u"Yellow House",
                           u"Post box # 3",
                           u"65789 LONDON (London County)",
                           u"UNITED KINGDOM" ]
        self.assertEqual(addr_lines, expected_lines)


# definitions for automatic unit testing

if __name__ == '__main__':
    unittest_main()



# -*- coding: utf-8 -*-
"""
``pybill.lib.entities.accounting_docs.abstract`` module is a private
module containing the abstract class that all the high-level entity classes
inherit from.

.. autoclass:: GenericAccountingDoc
   :members:
   :undoc-members:
   :show-inheritance:
"""
__docformat__ = "restructuredtext en"

from pybill.lib import ACCURACY, PBD_1_0


class GenericAccountingDoc:
    """
    Abstract class containing common parts for ``Bill``, ``Downpayment``,
    ``ClaimForm``, ``ProForma`` and ``Debit``.

    .. attribute:: id

       Attribute defiing the iIdentifier of the accounting document. This
       attribute is legally mandatory on some documents (e.g. bills that must
       have a unique identifier). On some other documents (e.g. pro-forma, it
       doesn't make sense). Thus, can be empty for some documents.

       type: :class:`unicode`

    .. attribute:: origin_format

       Attribute containing the name of the format from which this entity was
       created.  This attribute is set by the XML reader when the entity is
       created from a file. Can be ``u"PBD-1.0"`` or ``u"PBD-0.X"``. By
       default, it is set to the latest format (i.e. ``"PBD-1.0"``).
                         
       type: :class:`unicode`

    .. attribute:: doc_ref

       Attribute defining the reference of the document. It is different from
       the identifier because the document reference is actually used for
       classifying documents in the organisation whereas the identifier is for
       legal reference. Can be empty.

       type: :class:`unicode`

    .. attribute:: date

       Attribute defining the date when the document was published; this is a
       **string representation** of the date. The ``date_num`` attribute
       contains a `date` object. There is no correlation between these two
       attributes but they often refer to the same date. Can be empty.

       type: :class:`unicode`

    .. attribute:: date_num

       Attribute defining the date when the document was published; This is a
       :class:`~datetime.date` object. The ``date`` attribute contains a string
       representation of the date. There is no correlation between these two
       attributes but they often refer to the same date. Can be None.

       type: :class:`datetime.date`

    .. attribute:: place

       Attribute defining the place where the document was written. Can be
       empty.

       type: :class:`unicode`

    .. attribute:: other_infos

       Attribute defining a dictionary whose keys are the keywords of additional
       pieces of metadata and whose values are the values associated to these
       keywords for the document. For example: u"Supplier" and the supplier
       reference.

       type: dictionary of :class:`unicode` indexed by :class:`unicode` keys

    .. attribute:: sender

       Attribute containing the identification and various information about the
       sender of the document. These pieces of information complement the
       address of the organisation that emits the document (that is read from
       the configuration file). They actually consist in the sender name, its
       job titles, its phone and email.

       type: :class:`~pybill.lib.entities.addresses.PersonAddress`

    .. attribute:: receiver

       Attribute containing the name and address of the receiver of the
       document. The name of the person is used to display the person who should
       process the document. The organisation associated to this person is used
       to know which organisation will receive (and pay) the accounting document
       (e.g. bill).

       type: :class:`~pybill.lib.entities.addresses.PersonAddress`

    .. attribute:: acc_items

       Attribute containing the list of the items that appear on the accounting
       document (e.g. items sold that appear on a bill).
                     
       type: list of :class:`~pybill.lib.entities.accounting_docs.utils.AccItem`

    .. attribute:: remarks

       Attribute containing the optional remarks. Each remark is a text
       paragraph that is displayed at the beginning of the accounting document.

       type: list of :class:`unicode`

    .. attribute:: cfg_data

       Attribute containing the configuration data that must be used to render
       the accounting document. This configuration object is attached to the
       acccounting document by the XML reader.

       type: :class:`~pybill.lib.config.entities.ConfigData`

    .. attribute:: account_numbers

       Attribute containing the numbers of the accounting accounts that should
       be impacted by this accounting document. The numbers are stored in a
       dictionary whose keys are the ``CLIENT``, ``HOLDBACK_CLIENT``,
       ``PRODUCT`` and ``VAT`` constants.

       type: dictionary of :class:`unicode` indexed by :class:`unicode` keys

    .. automethod:: __init__
    """
    def __init__(self):
        """
        Initializes common parts of the object.

        This method raises a ``NotImplementedError`` exception if you try to
        instantiate an object of this class (that is abstract)
        """
        if self.__class__ is GenericAccountingDoc:
            raise NotImplementedError("Abstract classes can't be instanciated")
        # document standard metadata
        self.id = u""
        self.origin_format = PBD_1_0
        self.doc_ref = u""
        self.date = u""
        self.date_num = None
        self.place = u""
        # document other metadata
        self.other_infos = {}
        # sender and receiver
        self.sender = None
        self.receiver = None
        # List of items
        self.acc_items = []
        # Other text
        self.remarks = []
        # Config object attached to the document
        self.cfg_data = None
        # Accounting information
        self.account_numbers = {}

    def has_vat_rates(self):
        """
        This methods tests if the accounting document contains items with a
        VAT rate.
        
        :returns: True if the accounting document contains at least one item
                  with a VAT rate.
        :rtype: :class:`bool`
        """
        return reduce(lambda x, y: x or y,
                      [(it.vat_rate is not None) for it in self.acc_items],
                      False)

    def has_holdbacks(self):
        """
        This methods tests if the accounting document contains items with a
        holdback rate.
        
        :returns: True if the accounting document contains at least one item
                  with a holdback rate.
        :rtype: :class:`bool`
        """
        return reduce(lambda x, y: x or y,
                     [(it.holdback_rate is not None) for it in self.acc_items],
                      False)

    def has_vat_holdbacks(self):
        """
        This methods tests if the accounting document contains items with a
        holdback rate applied on VAT.
        
        :returns: True if the accounting document contains at least one item
                  with a holdback rate that is applied on VAT amount and this
                  item has a VAT rate.
        :rtype: :class:`bool`
        """
        return reduce(lambda x, y: x or y,
                      [(it.holdback_rate is not None
                        and it.holdback_on_vat
                        and it.vat_rate is not None) for it in self.acc_items],
                      False)

    def get_all_vat_rates(self):
        """
        Returns a set of the different VAT rates applied to the accounting
        items.

        Of course, each VAT rate appears only one time in the set. The set can
        be empty if no VAT is applied to the items.
        
        :returns: Set containing the different VAT rates that are applied to the
                  accounting items of the document
        :rtype: set of :class:`float`
        """
        vat_rates = set()
        for item in [i for i in self.acc_items if i.vat_rate is not None]:
            vat_rates.add(item.vat_rate)
        return vat_rates

    def get_vat_amount_for_rate(self, vat_rate):
        """
        Computes the sum of the VAT amounts of all the accounting items that use
        the given VAT rate.

        Returns ``0.0`` if no item matches the given VAT rate.
        
        :parameter vat_rate: VAT rate. Only the items that use this VAT rate
                             will be used to compute the VAT amount.
        :type vat_rate: :class:`float`
        
        :returns: VAT amount of all the items that use the specified VAT rate.
        :rtype: :class:`float`
        """
        vat_amount = 0.0
        for item in self.acc_items:
            if item.vat_rate == vat_rate:
                vat_amount += item.get_vat_amount()
        return round(vat_amount, ACCURACY)

    def get_price_total(self):
        """
        Computes the sum of the prices (actually tax free price) for all the
        items in the accounting document.
        
        :returns: Sum of the price of all the items.
        :rtype: :class:`float`
        """
        price_total = 0.0
        for item in self.acc_items:
            price_total += item.get_price()
        return round(price_total, ACCURACY)

    def get_holdback_amount(self):
        """
        Computes the sum of the holdbacks of all the accounting items.

        Returns ``0.0`` if no item has a holdback. Please note that these
        holdbacks are on tax free prices, see method ``get_holdback_vat_amount``
        method for holdbacks on VAT amount.
        
        :returns: Sum of the holdbacks for all the items in the accounting
                  document.
        :rtype: :class:`float`
        """
        holdback_amount = 0.0
        for item in self.acc_items:
            holdback_amount += item.get_holdback_amount()
        return round(holdback_amount, ACCURACY)

    def get_holdback_vat_amount(self):
        """
        Computes the sum of the holdbacks on VAT for all the accounting items.
        
        Returns ``0.0`` if no item has a holdback or no item has a VAT rate.
        Please note that these holdbacks are on VAT amounts, see method
        ``get_holdback_amount`` method for holdbacks on tax free prices.
        
        :returns: Sum of the holdbacks on VAT amounts for all the items in the
                  accounting document.
        :rtype: :class:`float`
        """
        holdback_vat_amount = 0.0
        for item in self.acc_items:
            holdback_vat_amount += item.get_holdback_vat_amount()
        return round(holdback_vat_amount, ACCURACY)

    def get_including_taxes_total(self):
        """
        Computes the sum of the including taxes prices for all the items in the
        accounting document.

        If the document contains only items with no VAT rate defined, this total
        is identical to the total returned by ``get_price_total`` method.
        
        :returns: Sum of the including taxes prices of all the items.
        :rtype: :class:`float`
        """
        including_taxes_total = 0.0
        for item in self.acc_items:
            including_taxes_total += item.get_including_taxes_price()
        return round(including_taxes_total, ACCURACY)

    def get_to_be_paid_holdback_free(self):
        """
        Computes the total without holdbacks that must be paid.

        This abstract method will be defined in the child classes.

        :returns: The total without holdbacks that must be paid. The
                  computation of this total depends on the accounting document
                  type.
        :rtype: :class:`float`
        """
        raise NotImplementedError()

    def get_to_be_paid_holdback(self):
        """
        Computes the total of holdbacks that must be paid.

        This abstract method will be defined in the child classes.

        :returns: The total of holdbacks that must be paid. The computation of
                  this total depends on the accounting document type.
        :rtype: :class:`float`
        """
        raise NotImplementedError()

    def get_to_be_paid_vat(self):
        """
        Computes the total of VAT amounts that must be paid.

        This abstract method will be defined in the child classes.

        :returns: The total of VAT amounts that must be paid. The computation 
                  of this total depends on the accounting document type.
        :rtype: :class:`float`
        """
        raise NotImplementedError()

    def get_to_be_paid(self):
        """
        Computes the total that must be paid.

        This method can be redefined in the child classes.

        Actually it is::

            total without holdbacks to be paid + total of holdbacks to be paid

        :returns: The total that must be paid. The computation of this total
                  depends on the accounting document type.
        :rtype: :class:`float`
        """
        return round(self.get_to_be_paid_holdback_free()
                     + self.get_to_be_paid_holdback(), ACCURACY)



# -*- coding: utf-8 -*-

from logilab.common.testlib import TestCase, unittest_main

import datetime
from StringIO import StringIO
from pybill.lib.accounting.utils import Entry
from pybill.lib.accounting.generator import EntriesGenerator

EXPECTED_XML = """<?xml version="1.0" encoding="%s"?>

<ecritures>
  <ecriture date="2009-11-06">
    <libelle>ENTRY 1</libelle>
    <debit compte="4111" montant="56.0"/>
    <debit compte="4111" montant="23.45"/>
    <debit compte="4112" montant="12.34"/>
    <credit compte="4451" montant="11.79"/>
    <credit compte="701" montant="80.0"/>
  </ecriture>
  <ecriture date="2009-11-06">
    <libelle>ENTRY 2</libelle>
    <debit compte="4111" montant="123.0"/>
    <debit compte="4113" montant="45.6"/>
    <credit compte="4111" montant="77.59"/>
    <credit compte="701" montant="91.01"/>
  </ecriture>
</ecritures>
"""


class EntriesGeneratorTest(TestCase):
    def setUp(self):
        """
        Called before each test from this class
        """
        self.generator = EntriesGenerator()
        ent = Entry(datetime.date(2009, 11, 06), u"ENTRY 1")
        ent.debits = {u"4111": [56.00, 23.45], u"4112": [12.34]}
        ent.credits = {u"701": [80.00], u"4451": [11.79]}
        self.generator.entries.append(ent)
        ent = Entry(datetime.date(2009, 11, 06), u"ENTRY 2")
        ent.debits = {u"4111": [123.00], u"4113": [45.60]}
        ent.credits = {u"701": [91.01], u"4111": [77.59]}
        self.generator.entries.append(ent)

    def test_xml_writing(self):
        out = StringIO()
        self.generator.write_entries_in_xml(out)
        self.assertMultiLineEqual(out.getvalue(), EXPECTED_XML % "UTF-8")

    def test_xml_writing_other_encoding(self):
        out = StringIO()
        self.generator.write_entries_in_xml(out, encoding="ISO-8859-1")
        self.assertMultiLineEqual(out.getvalue(), EXPECTED_XML % "ISO-8859-1")

    def test_xml_writing_no_entry(self):
        out = StringIO()
        self.generator.reset_entries()
        self.generator.write_entries_in_xml(out)
        self.assertMultiLineEqual(out.getvalue(), "")


# definitions for automatic unit testing

if __name__ == '__main__':
    unittest_main()



# -*- coding: utf-8 -*-
"""
``pybill.lib.xmlwriters.accdoc`` module contains the function used for saving
PyBill accounting documents into XML files.

The XML files are built thanks to :mod:`lxml` library. All the type of
accountind documents (bill, claim-form, debit, downpayment, pro-forma) can be
saved by this function.

Currently, it is only possible to save the entities in the latest
`PyBill Document 1.0` format.

.. autofunction:: write_accounting_doc
"""
__docformat__ = "restructuredtext en"

from lxml import etree

from pybill.lib import PBD_1_0
from pybill.lib.errors import PyBillWritingException

from pybill.lib.xmlwriters.utils import ENCODING, open_xml_outfile
from pybill.lib.xmlwriters.accdoc_format_1_0 import write_accdoc_1_0_xml


def write_accounting_doc(acc_doc, xml_file, format=PBD_1_0, encoding=ENCODING):
    """
    Writes an accounting entity into an XML file in ``format`` PyBill format.

    If the entity is linked to a configuration object that has a reference, a
    processing instruction will be added in the XML in order to specify
    the configuration to be used.

    This function relies on the other functions defined in this package that
    know how to write the XML corresponding to a given format.

    :parameter acc_doc: PyBill entity describing the accounting document to
                        be saved in XML. Can be a ``Bill``, a ``ClaimForm``,
                        a ``Debit``, a ``Downpayment`` or a ``ProForma``
                        object.
    :type acc_doc: :class:`~pybill.lib.entities.accounting_docs.abstract.GenericAccountingDoc`

    :parameter xml_file: Name of the file where the XML has to be saved.
    :type xml_file: :class:`unicode`

    :parameter format: name of the PyBill format used for the XML. Currently,
                       only `PyBill Document 1.0` format is supported; therefore
                       the only possible value for this parameter is
                       ``u"PBD-1.0"``.
    :type format: :class:`unicode`

    :parameter encoding: Encoding used to write the XML file.
    :type encoding: :class:`str`
    """
    if format == PBD_1_0:
        root_elt = write_accdoc_1_0_xml(acc_doc)
    else:
        raise PyBillWritingException("Format '%s' is not currently supported "
                                     "by the XML writer." % format)        
    out = open_xml_outfile(xml_file, encoding=encoding)
    # If the accounting document is linked to a configuration, writes a
    # processing instruction in the XML file.
    if (acc_doc.cfg_data is not None and acc_doc.cfg_data.name):
        cfg_pi = etree.ProcessingInstruction(u"pybill",
                                        u'config="%s"' % acc_doc.cfg_data.name)
        out.write(etree.tostring(cfg_pi, encoding=encoding, 
                                 pretty_print=True, xml_declaration=False))
        out.write("\n")
    # Finally, writes the XML elements.
    out.write(etree.tostring(root_elt, encoding=encoding, 
                             pretty_print=True, xml_declaration=False))
